# SPDX-License-Identifier: LGPL-2.1-or-later

# ***************************************************************************
# *   Copyright (c) 2015 Yorik van Havre <yorik@uncreated.net>              *
# *                                                                         *
# *   This program is free software; you can redistribute it and/or modify  *
# *   it under the terms of the GNU Lesser General Public License (LGPL)    *
# *   as published by the Free Software Foundation; either version 2 of     *
# *   the License, or (at your option) any later version.                   *
# *   for detail see the LICENCE text file.                                 *
# *                                                                         *
# *   This program is distributed in the hope that it will be useful,       *
# *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
# *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
# *   GNU Library General Public License for more details.                  *
# *                                                                         *
# *   You should have received a copy of the GNU Library General Public     *
# *   License along with this program; if not, write to the Free Software   *
# *   Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  *
# *   USA                                                                   *
# *                                                                         *
# ***************************************************************************

import FreeCAD
import FreeCADGui
import Path
import Path.Op.Base as PathOp
import PathScripts.PathUtils as PathUtils
import Path.Base.Util as PathUtil
from Path.Dressup.Utils import toolController
from PySide import QtCore

import random

from PySide.QtCore import QT_TRANSLATE_NOOP

__doc__ = """CAM Array object and FreeCAD command"""

translate = FreeCAD.Qt.translate


class ObjectArray:
    def __init__(self, obj):
        obj.addProperty(
            "App::PropertyLinkList",
            "Base",
            "Path",
            QT_TRANSLATE_NOOP("App::Property", "The toolpaths to array"),
        )
        obj.addProperty(
            "App::PropertyEnumeration",
            "Type",
            "Path",
            QT_TRANSLATE_NOOP("App::Property", "Pattern method"),
        )
        obj.addProperty(
            "App::PropertyVectorDistance",
            "Offset",
            "Path",
            QT_TRANSLATE_NOOP(
                "App::Property",
                "The spacing between the array copies in linear pattern",
            ),
        )
        obj.addProperty(
            "App::PropertyInteger",
            "CopiesX",
            "Path",
            QT_TRANSLATE_NOOP(
                "App::Property", "The number of copies in X-direction in linear pattern"
            ),
        )
        obj.addProperty(
            "App::PropertyInteger",
            "CopiesY",
            "Path",
            QT_TRANSLATE_NOOP(
                "App::Property", "The number of copies in Y-direction in linear pattern"
            ),
        )
        obj.addProperty(
            "App::PropertyAngle",
            "Angle",
            "Path",
            QT_TRANSLATE_NOOP("App::Property", "Total angle in polar pattern"),
        )
        obj.addProperty(
            "App::PropertyInteger",
            "Copies",
            "Path",
            QT_TRANSLATE_NOOP(
                "App::Property", "The number of copies in linear 1D and polar pattern"
            ),
        )
        obj.addProperty(
            "App::PropertyVector",
            "Centre",
            "Path",
            QT_TRANSLATE_NOOP("App::Property", "The centre of rotation in polar pattern"),
        )
        obj.addProperty(
            "App::PropertyBool",
            "SwapDirection",
            "Path",
            QT_TRANSLATE_NOOP(
                "App::Property",
                "Make copies in X direction before Y in Linear 2D pattern",
            ),
        )
        obj.addProperty(
            "App::PropertyInteger",
            "JitterPercent",
            "Path",
            QT_TRANSLATE_NOOP("App::Property", "Percent of copies to randomly offset"),
        )
        obj.addProperty(
            "App::PropertyVectorDistance",
            "JitterMagnitude",
            "Path",
            QT_TRANSLATE_NOOP("App::Property", "Maximum random offset of copies"),
        )
        obj.addProperty(
            "App::PropertyInteger",
            "JitterSeed",
            "Path",
            QT_TRANSLATE_NOOP("App::Property", "Seed value for jitter randomness"),
        )
        obj.addProperty(
            "App::PropertyLink",
            "ToolController",
            "Path",
            QT_TRANSLATE_NOOP(
                "App::Property",
                "The tool controller that will be used to calculate the toolpath",
            ),
        )
        obj.addProperty(
            "App::PropertyBool",
            "Active",
            "Path",
            QT_TRANSLATE_NOOP("PathOp", "Make False, to prevent operation from generating code"),
        )
        obj.addProperty(
            "App::PropertyString",
            "CycleTime",
            "Path",
            QT_TRANSLATE_NOOP("App::Property", "Operations cycle time estimation"),
        )

        obj.setEditorMode("CycleTime", 1)  # read-only
        obj.Active = True
        obj.Type = ["Linear1D", "Linear2D", "Polar"]

        self.setEditorModes(obj)
        obj.Proxy = self

    def dumps(self):
        return None

    def loads(self, state):
        return None

    def setEditorModes(self, obj):
        if obj.Type == "Linear1D":
            angleMode = centreMode = copiesXMode = copiesYMode = swapDirectionMode = 2
            copiesMode = offsetMode = 0
        elif obj.Type == "Linear2D":
            angleMode = copiesMode = centreMode = 2
            copiesXMode = copiesYMode = offsetMode = swapDirectionMode = 0
        elif obj.Type == "Polar":
            angleMode = copiesMode = centreMode = 0
            copiesXMode = copiesYMode = offsetMode = swapDirectionMode = 2

        obj.setEditorMode("Angle", angleMode)
        obj.setEditorMode("Copies", copiesMode)
        obj.setEditorMode("Centre", centreMode)
        obj.setEditorMode("CopiesX", copiesXMode)
        obj.setEditorMode("CopiesY", copiesYMode)
        obj.setEditorMode("Offset", offsetMode)
        obj.setEditorMode("SwapDirection", swapDirectionMode)
        obj.setEditorMode("JitterPercent", 0)
        obj.setEditorMode("JitterMagnitude", 0)
        obj.setEditorMode("JitterSeed", 0)
        obj.setEditorMode("ToolController", 2)

    def onChanged(self, obj, prop):
        if prop == "Type":
            self.setEditorModes(obj)

    def onDocumentRestored(self, obj):
        """onDocumentRestored(obj) ... Called automatically when document is restored."""

        if not hasattr(obj, "Active"):
            obj.addProperty(
                "App::PropertyBool",
                "Active",
                "Path",
                QT_TRANSLATE_NOOP(
                    "PathOp", "Make False, to prevent operation from generating code"
                ),
            )
            obj.Active = True

        if not hasattr(obj, "JitterSeed"):
            obj.addProperty(
                "App::PropertyInteger",
                "JitterSeed",
                "Path",
                QtCore.QT_TRANSLATE_NOOP("App::Property", "Seed value for jitter randomness"),
            )
            obj.JitterSeed = 0

        if not hasattr(obj, "CycleTime"):
            obj.addProperty(
                "App::PropertyString",
                "CycleTime",
                "Path",
                QT_TRANSLATE_NOOP("App::Property", "Operations cycle time estimation"),
            )
            obj.CycleTime = self.getCycleTimeEstimate(obj)

        self.setEditorModes(obj)

    def execute(self, obj):
        # backwards compatibility for PathArrays created before support for multiple bases
        if isinstance(obj.Base, list):
            base = obj.Base
        else:
            base = [obj.Base]

        # Do not generate paths and clear current Path data
        # if operation not Active or no base operations or operations not compatible
        if not obj.Active or len(base) == 0 or not self.isBaseCompatible(obj):
            obj.Path = Path.Path()
            return

        obj.ToolController = toolController(base[0])

        # use seed if specified, otherwise default to object name for consistency during recomputes
        seed = obj.JitterSeed or obj.Name

        pa = PathArray(
            obj.Base,
            obj.Type,
            obj.Copies,
            obj.Offset,
            obj.CopiesX,
            obj.CopiesY,
            obj.Angle,
            obj.Centre,
            obj.SwapDirection,
            obj.JitterMagnitude,
            obj.JitterPercent,
            seed,
        )

        obj.Path = pa.getPath()
        obj.CycleTime = PathOp.getCycleTimeEstimate(obj)

    def isBaseCompatible(self, obj):
        if not obj.Base:
            return False
        tcs = []
        cms = []
        for sel in obj.Base:
            if not sel.isDerivedFrom("Path::Feature"):
                return False
            tcs.append(toolController(sel))
            cms.append(PathUtil.coolantModeForOp(sel))

        if tcs == {None} or len(set(tcs)) > 1:
            Path.Log.warning(
                translate(
                    "PathArray",
                    "Arrays of toolpaths having different tool controllers or tool controller not selected.",
                )
            )
            return False

        if set(cms) != {"None"}:
            Path.Log.warning(
                translate(
                    "PathArray",
                    "Arrays not compatible with coolant modes.",
                )
            )
            return False

        return True


class PathArray:
    """class PathArray ...
    This class receives one or more base operations and repeats those operations
    at set intervals based upon array type requested and the related settings for that type."""

    def __init__(
        self,
        baseList,
        arrayType,
        copies,
        offsetVector,
        copiesX,
        copiesY,
        angle,
        centre,
        swapDirection,
        jitterMagnitude=FreeCAD.Vector(0, 0, 0),
        jitterPercent=0,
        seed="FreeCAD",
    ):
        self.baseList = list()
        self.arrayType = arrayType  # ['Linear1D', 'Linear2D', 'Polar']
        self.copies = copies
        self.offsetVector = offsetVector
        self.copiesX = copiesX
        self.copiesY = copiesY
        self.angle = angle
        self.centre = centre
        self.swapDirection = swapDirection
        self.jitterMagnitude = jitterMagnitude
        self.jitterPercent = jitterPercent
        self.seed = seed

        if baseList:
            if isinstance(baseList, list):
                self.baseList = baseList
            else:
                self.baseList = [baseList]

    # Private method
    def _calculateJitter(self, pos):
        """_calculateJitter(pos) ...
        Returns the position argument with a random vector shift applied."""
        if self.jitterPercent == 0:
            pass
        elif random.randint(0, 100) < self.jitterPercent:
            pos.x = pos.x + random.uniform(-self.jitterMagnitude.x, self.jitterMagnitude.x)
            pos.y = pos.y + random.uniform(-self.jitterMagnitude.y, self.jitterMagnitude.y)
            pos.z = pos.z + random.uniform(-self.jitterMagnitude.z, self.jitterMagnitude.z)
        return pos

    # Public method
    def getPath(self):
        """getPath() ... Call this method on an instance of the class to generate and return
        path data for the requested path array."""

        base = self.baseList
        for b in base:
            if not b.isDerivedFrom("Path::Feature"):
                return
            if not b.Path:
                return

            b_tool_controller = toolController(b)
            if not b_tool_controller:
                return

        # build copies
        output = ""
        random.seed(self.seed)

        if self.arrayType == "Linear1D":
            for i in range(self.copies):
                pos = FreeCAD.Vector(
                    self.offsetVector.x * (i + 1),
                    self.offsetVector.y * (i + 1),
                    self.offsetVector.z * (i + 1),
                )
                pos = self._calculateJitter(pos)

                for b in base:
                    pl = FreeCAD.Placement()
                    pl.move(pos)
                    np = Path.Path(
                        [cm.transform(pl) for cm in PathUtils.getPathWithPlacement(b).Commands]
                    )
                    output += np.toGCode()

        elif self.arrayType == "Linear2D":
            if self.swapDirection:
                for i in range(self.copiesY + 1):
                    for j in range(self.copiesX + 1):
                        if (i % 2) == 0:
                            pos = FreeCAD.Vector(
                                self.offsetVector.x * j,
                                self.offsetVector.y * i,
                                self.offsetVector.z * i,
                            )
                        else:
                            pos = FreeCAD.Vector(
                                self.offsetVector.x * (self.copiesX - j),
                                self.offsetVector.y * i,
                                self.offsetVector.z * i,
                            )
                        pos = self._calculateJitter(pos)

                        for b in base:
                            pl = FreeCAD.Placement()
                            # do not process the index 0,0. It will be processed by the base Paths themselves
                            if not (i == 0 and j == 0):
                                pl.move(pos)
                                np = Path.Path(
                                    [
                                        cm.transform(pl)
                                        for cm in PathUtils.getPathWithPlacement(b).Commands
                                    ]
                                )
                                output += np.toGCode()
            else:
                for i in range(self.copiesX + 1):
                    for j in range(self.copiesY + 1):
                        if (i % 2) == 0:
                            pos = FreeCAD.Vector(
                                self.offsetVector.x * i,
                                self.offsetVector.y * j,
                                self.offsetVector.z * i,
                            )
                        else:
                            pos = FreeCAD.Vector(
                                self.offsetVector.x * i,
                                self.offsetVector.y * (self.copiesY - j),
                                self.offsetVector.z * i,
                            )
                        pos = self._calculateJitter(pos)

                        for b in base:
                            pl = FreeCAD.Placement()
                            # do not process the index 0,0. It will be processed by the base Paths themselves
                            if not (i == 0 and j == 0):
                                pl.move(pos)
                                np = Path.Path(
                                    [
                                        cm.transform(pl)
                                        for cm in PathUtils.getPathWithPlacement(b).Commands
                                    ]
                                )
                                output += np.toGCode()
            # Eif
        else:
            for i in range(self.copies):
                for b in base:
                    ang = 360
                    if self.copies > 0:
                        ang = self.angle / self.copies * (1 + i)

                    pl = FreeCAD.Placement()
                    pl.rotate(self.centre, FreeCAD.Vector(0, 0, 1), ang)
                    np = PathUtils.applyPlacementToPath(pl, PathUtils.getPathWithPlacement(b))
                    output += np.toGCode()

        # return output
        return Path.Path(output)


class ViewProviderArray:
    def __init__(self, vobj):
        self.Object = vobj.Object
        vobj.Proxy = self

    def attach(self, vobj):
        self.Object = vobj.Object
        return

    def dumps(self):
        return None

    def loads(self, state):
        return None

    def claimChildren(self):
        if hasattr(self, "Object"):
            if hasattr(self.Object, "Base"):
                if self.Object.Base:
                    return self.Object.Base
        return []


class CommandPathArray:
    def GetResources(self):
        return {
            "Pixmap": "CAM_Array",
            "MenuText": QT_TRANSLATE_NOOP("CAM_Array", "Array"),
            "ToolTip": QT_TRANSLATE_NOOP("CAM_Array", "Creates an array from selected toolpaths"),
        }

    def IsActive(self):
        selection = FreeCADGui.Selection.getSelection()
        if not selection:
            return False
        tcs = []
        for sel in selection:
            if not sel.isDerivedFrom("Path::Feature"):
                return False
            tc = toolController(sel)
            if tc:
                # Active only for operations with identical tool controller
                tcs.append(tc)
                if len(set(tcs)) != 1:
                    return False
            else:
                return False
            if PathUtil.coolantModeForOp(sel) != "None":
                # Active only for operations without cooling
                return False
        return True

    def Activated(self):

        # check that the selection contains exactly what we want
        selection = FreeCADGui.Selection.getSelection()

        for sel in selection:
            if not (sel.isDerivedFrom("Path::Feature")):
                FreeCAD.Console.PrintError(
                    translate("CAM_Array", "Arrays can be created only from toolpath operations.")
                    + "\n"
                )
                return

        # if everything is ok, execute and register the transaction in the
        # undo/redo stack
        FreeCAD.ActiveDocument.openTransaction("Create Array")
        FreeCADGui.addModule("Path.Op.Gui.Array")
        FreeCADGui.addModule("PathScripts.PathUtils")

        FreeCADGui.doCommand(
            'obj = FreeCAD.ActiveDocument.addObject("Path::FeaturePython","Array")'
        )

        FreeCADGui.doCommand("Path.Op.Gui.Array.ObjectArray(obj)")

        baseString = "[%s]" % ",".join(
            ["FreeCAD.ActiveDocument.%s" % sel.Name for sel in selection]
        )
        FreeCADGui.doCommand("obj.Base = %s" % baseString)

        FreeCADGui.doCommand("obj.ViewObject.Proxy = 0")
        FreeCADGui.doCommand("PathScripts.PathUtils.addToJob(obj)")
        FreeCAD.ActiveDocument.commitTransaction()
        FreeCAD.ActiveDocument.recompute()


if FreeCAD.GuiUp:
    # register the FreeCAD command
    FreeCADGui.addCommand("CAM_Array", CommandPathArray())



# SPDX-License-Identifier: LGPL-2.1-or-later

# ***************************************************************************
# *                                                                         *
# *   Copyright (c) 2011 Yorik van Havre <yorik@uncreated.net>              *
# *                                                                         *
# *   This file is part of FreeCAD.                                         *
# *                                                                         *
# *   FreeCAD is free software: you can redistribute it and/or modify it    *
# *   under the terms of the GNU Lesser General Public License as           *
# *   published by the Free Software Foundation, either version 2.1 of the  *
# *   License, or (at your option) any later version.                       *
# *                                                                         *
# *   FreeCAD is distributed in the hope that it will be useful, but        *
# *   WITHOUT ANY WARRANTY; without even the implied warranty of            *
# *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU      *
# *   Lesser General Public License for more details.                       *
# *                                                                         *
# *   You should have received a copy of the GNU Lesser General Public      *
# *   License along with FreeCAD. If not, see                               *
# *   <https://www.gnu.org/licenses/>.                                      *
# *                                                                         *
# ***************************************************************************

__title__ = "FreeCAD Axis System"
__author__ = "Yorik van Havre"
__url__ = "https://www.freecad.org"

## @package ArchAxis
#  \ingroup ARCH
#  \brief Axis for the Arch workbench
#
#  This module provides tools to build axis
#  An axis is a collection of planar axes with a number/tag

import math

import FreeCAD
import ArchCommands
import Draft
import Part

from FreeCAD import Vector
from draftutils import params

if FreeCAD.GuiUp:
    import re
    from pivy import coin
    from PySide import QtCore, QtGui
    from PySide.QtCore import QT_TRANSLATE_NOOP
    import FreeCADGui
    from draftutils.translate import translate
else:
    # \cond
    def translate(ctxt, txt):
        return txt

    def QT_TRANSLATE_NOOP(ctxt, txt):
        return txt

    # \endcond


class _Axis:
    "The Axis object"

    def __init__(self, obj):

        obj.Proxy = self
        self.Type = "Axis"
        self.setProperties(obj)

    def setProperties(self, obj):

        pl = obj.PropertiesList
        if not "Distances" in pl:
            obj.addProperty(
                "App::PropertyFloatList",
                "Distances",
                "Axis",
                QT_TRANSLATE_NOOP("App::Property", "The intervals between axes"),
                locked=True,
            )
        if not "Angles" in pl:
            obj.addProperty(
                "App::PropertyFloatList",
                "Angles",
                "Axis",
                QT_TRANSLATE_NOOP("App::Property", "The angles of each axis"),
                locked=True,
            )
        if not "Labels" in pl:
            obj.addProperty(
                "App::PropertyStringList",
                "Labels",
                "Axis",
                QT_TRANSLATE_NOOP("App::Property", "The label of each axis"),
                locked=True,
            )
        if not "CustomNumber" in pl:
            obj.addProperty(
                "App::PropertyString",
                "CustomNumber",
                "Axis",
                QT_TRANSLATE_NOOP("App::Property", "An optional custom bubble number"),
                locked=True,
            )
        if not "Length" in pl:
            obj.addProperty(
                "App::PropertyLength",
                "Length",
                "Axis",
                QT_TRANSLATE_NOOP("App::Property", "The length of the axes"),
                locked=True,
            )
            obj.Length = 3000
        if not "Placement" in pl:
            obj.addProperty("App::PropertyPlacement", "Placement", "Base", "", locked=True)
        if not "Shape" in pl:
            obj.addProperty("Part::PropertyPartShape", "Shape", "Base", "", locked=True)
        if not "Limit" in pl:
            obj.addProperty(
                "App::PropertyLength",
                "Limit",
                "Axis",
                QT_TRANSLATE_NOOP(
                    "App::Property",
                    "If not zero, the axes are not represented as one full line but as two lines of the given length",
                ),
                locked=True,
            )
            obj.Limit = 0

    def onDocumentRestored(self, obj):

        self.setProperties(obj)

    def execute(self, obj):

        pl = obj.Placement
        geoms = []
        dist = 0
        distances = [0]
        angles = [0]
        if hasattr(obj, "Distances"):
            distances = obj.Distances
        if hasattr(obj, "Angles"):
            angles = obj.Angles
        if distances and obj.Length.Value:
            if angles and len(distances) == len(angles):
                for i in range(len(distances)):
                    dist += distances[i]
                    ang = math.radians(angles[i])
                    ln = obj.Length.Value
                    ln = 100 * ln if abs(math.cos(ang)) < 0.01 else ln / math.cos(ang)
                    unitvec = Vector(math.sin(ang), math.cos(ang), 0)
                    p1 = Vector(dist, 0, 0)
                    p2 = p1 + unitvec * ln
                    if hasattr(obj, "Limit") and obj.Limit.Value:
                        p3 = unitvec * obj.Limit.Value
                        p4 = unitvec * -obj.Limit.Value
                        geoms.append(Part.LineSegment(p1, p1 + p3).toShape())
                        geoms.append(Part.LineSegment(p2, p2 + p4).toShape())
                    else:
                        geoms.append(Part.LineSegment(p1, p2).toShape())
        if geoms:
            sh = Part.Compound(geoms)
            obj.Shape = sh
            obj.Placement = pl

    def onChanged(self, obj, prop):

        if prop in ["Angles", "Distances", "Placement"]:
            obj.touch()

    def dumps(self):

        return None

    def loads(self, state):

        self.Type = "Axis"

    def getPoints(self, obj):
        "returns the gridpoints of linked axes"

        pts = []
        for e in obj.Shape.Edges:
            pts.append(e.Vertexes[0].Point)
        return pts

    def getAxisData(self, obj):
        data = []
        num = 0
        for e in obj.Shape.Edges:
            axdata = []
            axdata.append(e.Vertexes[0].Point)
            axdata.append(e.Vertexes[-1].Point)
            if obj.ViewObject:
                axdata.append(obj.ViewObject.Proxy.getNumber(obj.ViewObject, num))
            else:
                axdata.append(str(num))
            data.append(axdata)
            num += 1
        return data


class _ViewProviderAxis:
    "A View Provider for the Axis object"

    def __init__(self, vobj):

        vobj.Proxy = self
        self.setProperties(vobj)

    def setProperties(self, vobj):

        ts = params.get_param("textheight") * params.get_param("DefaultAnnoScaleMultiplier")
        pl = vobj.PropertiesList
        if not "BubbleSize" in pl:
            vobj.addProperty(
                "App::PropertyLength",
                "BubbleSize",
                "Axis",
                QT_TRANSLATE_NOOP("App::Property", "The size of the axis bubbles"),
                locked=True,
            )
            vobj.BubbleSize = ts * 1.42
        if not "NumberingStyle" in pl:
            vobj.addProperty(
                "App::PropertyEnumeration",
                "NumberingStyle",
                "Axis",
                QT_TRANSLATE_NOOP("App::Property", "The numbering style"),
                locked=True,
            )
            vobj.NumberingStyle = [
                "1,2,3",
                "01,02,03",
                "001,002,003",
                "A,B,C",
                "a,b,c",
                "I,II,III",
                "L0,L1,L2",
            ]
            vobj.NumberingStyle = "1,2,3"
        if not "DrawStyle" in pl:
            vobj.addProperty(
                "App::PropertyEnumeration",
                "DrawStyle",
                "Axis",
                QT_TRANSLATE_NOOP("App::Property", "The type of line to draw this axis"),
                locked=True,
            )
            vobj.DrawStyle = ["Solid", "Dashed", "Dotted", "Dashdot"]
        vobj.DrawStyle = "Dashdot"
        if not "BubblePosition" in pl:
            vobj.addProperty(
                "App::PropertyEnumeration",
                "BubblePosition",
                "Axis",
                QT_TRANSLATE_NOOP(
                    "App::Property", "Where to add bubbles to this axis: Start, end, both or none"
                ),
                locked=True,
            )
            vobj.BubblePosition = [
                "Start",
                "End",
                "Both",
                "None",
                "Arrow left",
                "Arrow right",
                "Bar left",
                "Bar right",
            ]
        if not "LineWidth" in pl:
            vobj.addProperty(
                "App::PropertyFloat",
                "LineWidth",
                "Axis",
                QT_TRANSLATE_NOOP("App::Property", "The line width to draw this axis"),
                locked=True,
            )
            vobj.LineWidth = 1
        if not "LineColor" in pl:
            vobj.addProperty(
                "App::PropertyColor",
                "LineColor",
                "Axis",
                QT_TRANSLATE_NOOP("App::Property", "The color of this axis"),
                locked=True,
            )
        vobj.LineColor = ArchCommands.getDefaultColor("Helpers")
        if not "StartNumber" in pl:
            vobj.addProperty(
                "App::PropertyInteger",
                "StartNumber",
                "Axis",
                QT_TRANSLATE_NOOP("App::Property", "The number of the first axis"),
                locked=True,
            )
            vobj.StartNumber = 1
        if not "FontName" in pl:
            vobj.addProperty(
                "App::PropertyFont",
                "FontName",
                "Axis",
                QT_TRANSLATE_NOOP("App::Property", "The font to use for texts"),
                locked=True,
            )
            vobj.FontName = params.get_param("textfont")
        if not "FontSize" in pl:
            vobj.addProperty(
                "App::PropertyLength",
                "FontSize",
                "Axis",
                QT_TRANSLATE_NOOP("App::Property", "The font size"),
                locked=True,
            )
            vobj.FontSize = ts
        if not "ShowLabel" in pl:
            vobj.addProperty(
                "App::PropertyBool",
                "ShowLabel",
                "Axis",
                QT_TRANSLATE_NOOP("App::Property", "If true, show the labels"),
            )
        if not "LabelOffset" in pl:
            vobj.addProperty(
                "App::PropertyPlacement",
                "LabelOffset",
                "Axis",
                QT_TRANSLATE_NOOP("App::Property", "A transformation to apply to each label"),
                locked=True,
            )

    def onDocumentRestored(self, vobj):

        self.setProperties(vobj)

    def getIcon(self):

        import Arch_rc

        return ":/icons/Arch_Axis_Tree.svg"

    def claimChildren(self):

        return []

    def attach(self, vobj):
        self.Object = vobj.Object
        self.bubbles = None
        self.bubbletexts = []
        self.bubbledata = []
        sep = coin.SoSeparator()
        self.mat = coin.SoMaterial()
        self.linestyle = coin.SoDrawStyle()
        self.linecoords = coin.SoCoordinate3()
        self.lineset = coin.SoType.fromName("SoBrepEdgeSet").createInstance()
        self.bubbleset = coin.SoSeparator()
        self.labelset = coin.SoSeparator()
        sep.addChild(self.mat)
        sep.addChild(self.linestyle)
        sep.addChild(self.linecoords)
        sep.addChild(self.lineset)
        sep.addChild(self.bubbleset)
        sep.addChild(self.labelset)
        vobj.addDisplayMode(sep, "Default")
        self.onChanged(vobj, "BubbleSize")
        self.onChanged(vobj, "ShowLabel")
        self.onChanged(vobj, "LineColor")
        self.onChanged(vobj, "LineWidth")
        self.onChanged(vobj, "DrawStyle")

    def getDisplayModes(self, vobj):

        return ["Default"]

    def getDefaultDisplayMode(self):

        return "Default"

    def setDisplayMode(self, mode):

        return mode

    def updateData(self, obj, prop):

        if prop == "Shape":
            if obj.Shape:
                if obj.Shape.Edges:
                    verts = []
                    vset = []
                    i = 0
                    for e in obj.Shape.Edges:
                        for v in e.Vertexes:
                            verts.append(tuple(obj.Placement.inverse().multVec(v.Point)))
                            vset.append(i)
                            i += 1
                        vset.append(-1)
                    self.linecoords.point.setValues(verts)
                    self.lineset.coordIndex.setValues(0, len(vset), vset)
                    self.lineset.coordIndex.setNum(len(vset))
        elif prop in ["Placement", "Length"] and not hasattr(obj, "Distances"):
            # copy values from FlatLines/Wireframe nodes
            rn = obj.ViewObject.RootNode
            if rn.getNumChildren() < 3:
                return
            coords = rn.getChild(1)
            pts = coords.point.getValues()
            self.linecoords.point.setValues(pts)
            # self.linecoords.point.setNum(len(pts))
            sw = rn.getChild(2)
            if sw.getNumChildren() < 4:
                return
            edges = sw.getChild(sw.getNumChildren() - 2)
            if not edges.getNumChildren():
                return
            if edges.getChild(0).getNumChildren() < 4:
                return
            eset = edges.getChild(0).getChild(3)
            vset = eset.coordIndex.getValues()
            self.lineset.coordIndex.setValues(0, len(vset), vset)
            self.lineset.coordIndex.setNum(len(vset))
        self.onChanged(obj.ViewObject, "BubbleSize")
        self.onChanged(obj.ViewObject, "ShowLabel")

    def onChanged(self, vobj, prop):

        if prop == "LineColor":
            if hasattr(vobj, "LineColor"):
                l = vobj.LineColor
                self.mat.diffuseColor.setValue([l[0], l[1], l[2]])
        elif prop == "DrawStyle":
            if hasattr(vobj, "DrawStyle"):
                if vobj.DrawStyle == "Solid":
                    self.linestyle.linePattern = 0xFFFF
                elif vobj.DrawStyle == "Dashed":
                    self.linestyle.linePattern = 0xF00F
                elif vobj.DrawStyle == "Dotted":
                    self.linestyle.linePattern = 0x0F0F
                else:
                    self.linestyle.linePattern = 0xFF88
        elif prop == "LineWidth":
            if hasattr(vobj, "LineWidth"):
                self.linestyle.lineWidth = vobj.LineWidth
        elif prop in ["BubbleSize", "BubblePosition", "FontName", "FontSize"]:
            if hasattr(self, "bubbleset"):
                if self.bubbles:
                    self.bubbleset.removeChild(self.bubbles)
                    self.bubbles = None
                if vobj.Object.Shape:
                    if vobj.Object.Shape.Edges:
                        self.bubbles = coin.SoSeparator()
                        self.bubblestyle = coin.SoDrawStyle()
                        self.bubblestyle.linePattern = 0xFFFF
                        self.bubbles.addChild(self.bubblestyle)
                        self.bubbletexts = []
                        self.bubbledata = []
                        pos = ["Start"]
                        if hasattr(vobj, "BubblePosition"):
                            if vobj.BubblePosition in [
                                "Both",
                                "Arrow left",
                                "Arrow right",
                                "Bar left",
                                "Bar right",
                            ]:
                                pos = ["Start", "End"]
                            elif vobj.BubblePosition == "None":
                                pos = []
                            else:
                                pos = [vobj.BubblePosition]
                        e = len(vobj.Object.Shape.Edges)
                        if getattr(vobj.Object, "Limit", 0):
                            e //= 2
                        n = len(getattr(vobj.Object, "Distances", []))
                        for i in range(min(e, n)):
                            for p in pos:
                                if getattr(vobj.Object, "Limit", 0):
                                    verts = [
                                        vobj.Object.Placement.inverse().multVec(
                                            vobj.Object.Shape.Edges[i * 2].Vertexes[0].Point
                                        ),
                                        vobj.Object.Placement.inverse().multVec(
                                            vobj.Object.Shape.Edges[i * 2 + 1].Vertexes[0].Point
                                        ),
                                    ]
                                else:
                                    verts = [
                                        vobj.Object.Placement.inverse().multVec(v.Point)
                                        for v in vobj.Object.Shape.Edges[i].Vertexes
                                    ]
                                arrow = None
                                if p == "Start":
                                    p1 = verts[0]
                                    p2 = verts[1]
                                    if vobj.BubblePosition.endswith("left"):
                                        arrow = True
                                    elif vobj.BubblePosition.endswith("right"):
                                        arrow = False
                                else:
                                    p1 = verts[1]
                                    p2 = verts[0]
                                    if vobj.BubblePosition.endswith("left"):
                                        arrow = False
                                    elif vobj.BubblePosition.endswith("right"):
                                        arrow = True
                                dv = p2.sub(p1)
                                dv.normalize()
                                if hasattr(vobj.BubbleSize, "Value"):
                                    rad = vobj.BubbleSize.Value / 2
                                else:
                                    rad = vobj.BubbleSize / 2
                                center = p2.add(Vector(dv).multiply(rad))
                                normal = vobj.Object.Placement.Rotation.multVec(Vector(0, 0, 1))
                                chord = dv.cross(normal)
                                if arrow:
                                    p3 = p2.add(Vector(chord).multiply(rad / 2).negative())
                                    if vobj.BubblePosition.startswith("Arrow"):
                                        p4 = p3.add(Vector(dv).multiply(rad * 2).negative())
                                        p5 = p2.add(Vector(dv).multiply(rad).negative()).add(
                                            Vector(chord).multiply(rad * 1.5).negative()
                                        )
                                        pts = [tuple(p3), tuple(p5), tuple(p4), tuple(p3)]
                                        center = p5.add(Vector(chord).multiply(rad * 2.5))
                                    else:
                                        p4 = p3.add(Vector(dv).multiply(rad / 2).negative())
                                        p5 = p4.add(Vector(chord).multiply(rad * 1.5).negative())
                                        p6 = p5.add(Vector(dv).multiply(rad / 2))
                                        pts = [
                                            tuple(p3),
                                            tuple(p6),
                                            tuple(p5),
                                            tuple(p4),
                                            tuple(p3),
                                        ]
                                        center = p5.add(Vector(chord).multiply(rad * 3))
                                    coords = coin.SoCoordinate3()
                                    coords.point.setValues(0, len(pts), pts)
                                    line = coin.SoFaceSet()
                                    line.numVertices.setValue(-1)
                                    cir = Part.makePolygon(pts)
                                    cir.Placement = vobj.Object.Placement
                                    self.bubbledata.append(cir)
                                elif arrow == False:
                                    p3 = p2.add(Vector(chord).multiply(rad / 2))
                                    if vobj.BubblePosition.startswith("Arrow"):
                                        p4 = p3.add(Vector(dv).multiply(rad * 2).negative())
                                        p5 = p2.add(Vector(dv).multiply(rad).negative()).add(
                                            Vector(chord).multiply(rad * 1.5)
                                        )
                                        pts = [tuple(p3), tuple(p4), tuple(p5), tuple(p3)]
                                        center = p5.add(
                                            Vector(chord).multiply(rad * 2.5).negative()
                                        )
                                    else:
                                        p4 = p3.add(Vector(dv).multiply(rad / 2).negative())
                                        p5 = p4.add(Vector(chord).multiply(rad * 1.5))
                                        p6 = p5.add(Vector(dv).multiply(rad / 2))
                                        pts = [
                                            tuple(p3),
                                            tuple(p4),
                                            tuple(p5),
                                            tuple(p6),
                                            tuple(p3),
                                        ]
                                        center = p5.add(Vector(chord).multiply(rad * 3).negative())
                                    coords = coin.SoCoordinate3()
                                    coords.point.setValues(0, len(pts), pts)
                                    line = coin.SoFaceSet()
                                    line.numVertices.setValue(-1)
                                    cir = Part.makePolygon(pts)
                                    cir.Placement = vobj.Object.Placement
                                    self.bubbledata.append(cir)
                                else:
                                    cir = Part.makeCircle(rad, center)
                                    buf = cir.writeInventor()
                                    try:
                                        cin = coin.SoInput()
                                        cin.setBuffer(buf)
                                        cob = coin.SoDB.readAll(cin)
                                    except Exception:
                                        # workaround for pivy SoInput.setBuffer() bug
                                        buf = buf.replace("\n", "")
                                        pts = re.findall(r"point \[(.*?)\]", buf)[0]
                                        pts = pts.split(",")
                                        pc = []
                                        for point in pts:
                                            v = point.strip().split()
                                            pc.append([float(v[0]), float(v[1]), float(v[2])])
                                        coords = coin.SoCoordinate3()
                                        coords.point.setValues(0, len(pc), pc)
                                        line = coin.SoLineSet()
                                        line.numVertices.setValue(-1)
                                    else:
                                        coords = cob.getChild(1).getChild(0).getChild(2)
                                        line = cob.getChild(1).getChild(0).getChild(3)
                                    cir.Placement = vobj.Object.Placement
                                    self.bubbledata.append(cir)
                                self.bubbles.addChild(coords)
                                self.bubbles.addChild(line)
                                st = coin.SoSeparator()
                                tr = coin.SoTransform()
                                fs = rad * 1.5
                                if hasattr(vobj, "FontSize"):
                                    fs = vobj.FontSize.Value
                                txpos = FreeCAD.Vector(center.x, center.y - fs / 2.5, center.z)
                                tr.translation.setValue(tuple(txpos))
                                fo = coin.SoFont()
                                fn = params.get_param("textfont")
                                if hasattr(vobj, "FontName"):
                                    if vobj.FontName:
                                        try:
                                            fn = str(vobj.FontName)
                                        except Exception:
                                            pass
                                fo.name = fn
                                fo.size = fs
                                tx = coin.SoAsciiText()
                                tx.justification = coin.SoText2.CENTER
                                self.bubbletexts.append((tx, vobj.Object.Placement.multVec(center)))
                                st.addChild(tr)
                                st.addChild(fo)
                                st.addChild(tx)
                                self.bubbles.addChild(st)
                        self.bubbleset.addChild(self.bubbles)
                        self.onChanged(vobj, "NumberingStyle")
            if prop in ["FontName", "FontSize"]:
                self.onChanged(vobj, "ShowLabel")
        elif prop in ["NumberingStyle", "StartNumber"]:
            if hasattr(self, "bubbletexts"):
                num = 0
                if hasattr(vobj, "StartNumber"):
                    if vobj.StartNumber > 1:
                        num = vobj.StartNumber - 1
                alt = False
                for t in self.bubbletexts:
                    t[0].string = self.getNumber(vobj, num)
                    num += 1
                    if hasattr(vobj, "BubblePosition"):
                        if vobj.BubblePosition in [
                            "Both",
                            "Arrow left",
                            "Arrow right",
                            "Bar left",
                            "Bar right",
                        ]:
                            if not alt:
                                num -= 1
                    alt = not alt
        elif prop in ["ShowLabel", "LabelOffset"]:
            if hasattr(self, "labels"):
                if self.labels:
                    self.labelset.removeChild(self.labels)
            self.labels = None
            if hasattr(vobj, "ShowLabel") and hasattr(vobj.Object, "Labels"):
                if vobj.ShowLabel:
                    self.labels = coin.SoSeparator()
                    if hasattr(vobj.Object, "Limit") and vobj.Object.Limit.Value:
                        n = len(vobj.Object.Shape.Edges) / 2
                    else:
                        n = len(vobj.Object.Shape.Edges)
                    for i in range(n):
                        if len(vobj.Object.Labels) > i:
                            if vobj.Object.Labels[i]:
                                vert = vobj.Object.Shape.Edges[i].Vertexes[0].Point
                                if hasattr(vobj, "LabelOffset"):
                                    pl = FreeCAD.Placement(vobj.LabelOffset)
                                    pl.Base = vert.add(pl.Base)
                                st = coin.SoSeparator()
                                tr = coin.SoTransform()
                                fo = coin.SoFont()
                                tx = coin.SoAsciiText()
                                tx.justification = coin.SoText2.LEFT
                                t = vobj.Object.Labels[i]
                                tx.string.setValue(t)
                                if hasattr(vobj, "FontSize"):
                                    fs = vobj.FontSize.Value
                                elif hasattr(vobj.BubbleSize, "Value"):
                                    fs = vobj.BubbleSize.Value * 0.75
                                else:
                                    fs = vobj.BubbleSize * 0.75
                                tr.translation.setValue(tuple(pl.Base))
                                tr.rotation.setValue(pl.Rotation.Q)
                                fn = params.get_param("textfont")
                                if hasattr(vobj, "FontName"):
                                    if vobj.FontName:
                                        try:
                                            fn = str(vobj.FontName)
                                        except Exception:
                                            pass
                                fo.name = fn
                                fo.size = fs
                                st.addChild(tr)
                                st.addChild(fo)
                                st.addChild(tx)
                                self.labels.addChild(st)
                    self.labelset.addChild(self.labels)

    def getNumber(self, vobj, num):

        chars = "abcdefghijklmnopqrstuvwxyz"
        roman = (
            ("M", 1000),
            ("CM", 900),
            ("D", 500),
            ("CD", 400),
            ("C", 100),
            ("XC", 90),
            ("L", 50),
            ("XL", 40),
            ("X", 10),
            ("IX", 9),
            ("V", 5),
            ("IV", 4),
            ("I", 1),
        )
        if hasattr(vobj.Object, "CustomNumber") and vobj.Object.CustomNumber:
            return vobj.Object.CustomNumber
        elif hasattr(vobj, "NumberingStyle"):
            if vobj.NumberingStyle == "1,2,3":
                return str(num + 1)
            elif vobj.NumberingStyle == "01,02,03":
                return str(num + 1).zfill(2)
            elif vobj.NumberingStyle == "001,002,003":
                return str(num + 1).zfill(3)
            elif vobj.NumberingStyle == "A,B,C":
                result = ""
                base = num // 26
                if base:
                    result += chars[base].upper()
                remainder = num % 26
                result += chars[remainder].upper()
                return result
            elif vobj.NumberingStyle == "a,b,c":
                result = ""
                base = num // 26
                if base:
                    result += chars[base]
                remainder = num % 26
                result += chars[remainder]
                return result
            elif vobj.NumberingStyle == "I,II,III":
                result = ""
                n = num
                n += 1
                for numeral, integer in roman:
                    while n >= integer:
                        result += numeral
                        n -= integer
                return result
            elif vobj.NumberingStyle == "L0,L1,L2":
                return "L" + str(num)
        else:
            return str(num + 1)

    def getTextData(self):

        return [(t[0].string.getValues()[0], t[1]) for t in self.bubbletexts]

    def getShapeData(self):

        return self.bubbledata

    def setEdit(self, vobj, mode):
        if mode == 1 or mode == 2:
            return None

        taskd = _AxisTaskPanel()
        taskd.obj = vobj.Object
        taskd.update()
        FreeCADGui.Control.showDialog(taskd)
        return True

    def unsetEdit(self, vobj, mode):
        if mode == 1 or mode == 2:
            return None

        FreeCADGui.Control.closeDialog()
        return True

    def setupContextMenu(self, vobj, menu):
        if FreeCADGui.activeWorkbench().name() != "BIMWorkbench":
            return
        actionEdit = QtGui.QAction(translate("Arch", "Edit"), menu)
        QtCore.QObject.connect(actionEdit, QtCore.SIGNAL("triggered()"), self.edit)
        menu.addAction(actionEdit)

        # The default Part::FeaturePython context menu contains a `Set colors`
        # option. This option makes no sense for Axis objects. We therefore
        # override this menu and have to add our own `Transform` item.
        # To override the default menu this function must return `True`.
        action_transform = QtGui.QAction(
            FreeCADGui.getIcon("Std_TransformManip.svg"),
            translate("Command", "Transform"),  # Context `Command` instead of `Arch`.
            menu,
        )
        QtCore.QObject.connect(action_transform, QtCore.SIGNAL("triggered()"), self.transform)
        menu.addAction(action_transform)

        return True

    def edit(self):
        FreeCADGui.ActiveDocument.setEdit(self.Object, 0)

    def transform(self):
        FreeCADGui.ActiveDocument.setEdit(self.Object, 1)

    def dumps(self):

        return None

    def loads(self, state):

        return None


class _AxisTaskPanel:
    """The editmode TaskPanel for Axis objects"""

    def __init__(self):

        # the panel has a tree widget that contains categories
        # for the subcomponents, such as additions, subtractions.
        # the categories are shown only if they are not empty.

        self.updating = False

        self.obj = None
        self.form = QtGui.QWidget()
        self.form.setObjectName("TaskPanel")
        self.grid = QtGui.QGridLayout(self.form)
        self.grid.setObjectName("grid")
        self.title = QtGui.QLabel(self.form)
        self.grid.addWidget(self.title, 0, 0, 1, 2)

        # tree
        self.tree = QtGui.QTreeWidget(self.form)
        self.grid.addWidget(self.tree, 1, 0, 1, 2)
        self.tree.setColumnCount(4)
        self.tree.header().resizeSection(0, 50)
        self.tree.header().resizeSection(1, 80)
        self.tree.header().resizeSection(2, 60)

        # buttons
        self.addButton = QtGui.QPushButton(self.form)
        self.addButton.setObjectName("addButton")
        self.addButton.setIcon(QtGui.QIcon(":/icons/Arch_Add.svg"))
        self.grid.addWidget(self.addButton, 3, 0, 1, 1)
        self.addButton.setEnabled(True)

        self.delButton = QtGui.QPushButton(self.form)
        self.delButton.setObjectName("delButton")
        self.delButton.setIcon(QtGui.QIcon(":/icons/Arch_Remove.svg"))
        self.grid.addWidget(self.delButton, 3, 1, 1, 1)
        self.delButton.setEnabled(True)

        QtCore.QObject.connect(self.addButton, QtCore.SIGNAL("clicked()"), self.addElement)
        QtCore.QObject.connect(self.delButton, QtCore.SIGNAL("clicked()"), self.removeElement)
        QtCore.QObject.connect(
            self.tree, QtCore.SIGNAL("itemChanged(QTreeWidgetItem *, int)"), self.edit
        )
        self.update()

    def isAllowedAlterSelection(self):

        return False

    def isAllowedAlterView(self):

        return True

    def getStandardButtons(self):

        return QtGui.QDialogButtonBox.Close

    def update(self):
        "fills the treewidget"
        self.updating = True
        self.tree.clear()
        if self.obj and hasattr(self.obj, "Distances"):
            for i in range(len(self.obj.Distances)):
                item = QtGui.QTreeWidgetItem(self.tree)
                item.setText(0, str(i + 1))
                if len(self.obj.Distances) > i:
                    item.setText(1, str(self.obj.Distances[i]))
                if len(self.obj.Angles) > i:
                    item.setText(2, str(self.obj.Angles[i]))
                if hasattr(self.obj, "Labels"):
                    if len(self.obj.Labels) > i:
                        item.setText(3, str(self.obj.Labels[i]))
                item.setFlags(item.flags() | QtCore.Qt.ItemIsEditable)
                item.setTextAlignment(0, QtCore.Qt.AlignLeft)
        self.retranslateUi(self.form)
        self.updating = False

    def addElement(self):

        item = QtGui.QTreeWidgetItem(self.tree)
        item.setText(0, str(self.tree.topLevelItemCount()))
        item.setText(1, "1.0")
        item.setText(2, "0.0")
        item.setFlags(item.flags() | QtCore.Qt.ItemIsEditable)
        self.resetObject()

    def removeElement(self):

        it = self.tree.currentItem()
        if it:
            nr = int(it.text(0)) - 1
            self.resetObject(remove=nr)
            self.update()

    def edit(self, item, column):

        if not self.updating:
            self.resetObject()

    def resetObject(self, remove=None):
        "transfers the values from the widget to the object"

        d = []
        a = []
        l = []
        for i in range(self.tree.topLevelItemCount()):
            it = self.tree.findItems(str(i + 1), QtCore.Qt.MatchExactly, 0)[0]
            if (remove is None) or (remove != i):
                if it.text(1):
                    d.append(float(it.text(1)))
                else:
                    d.append(0.0)
                if it.text(2):
                    a.append(float(it.text(2)))
                else:
                    a.append(0.0)
                l.append(it.text(3))
        self.obj.Distances = d
        self.obj.Angles = a
        self.obj.Labels = l
        self.obj.touch()
        FreeCAD.ActiveDocument.recompute()

    def reject(self):

        FreeCAD.ActiveDocument.recompute()
        FreeCADGui.ActiveDocument.resetEdit()
        return True

    def retranslateUi(self, TaskPanel):

        TaskPanel.setWindowTitle(QtGui.QApplication.translate("Arch", "Axes", None))
        self.delButton.setText(QtGui.QApplication.translate("Arch", "Remove", None))
        self.addButton.setText(QtGui.QApplication.translate("Arch", "Add", None))
        self.title.setText(
            QtGui.QApplication.translate(
                "Arch", "Distances (mm) and angles (deg) between axes", None
            )
        )
        self.tree.setHeaderLabels(
            [
                QtGui.QApplication.translate("Arch", "Axis", None),
                QtGui.QApplication.translate("Arch", "Distance", None),
                QtGui.QApplication.translate("Arch", "Angle", None),
                QtGui.QApplication.translate("Arch", "Label", None),
            ]
        )



# SPDX-License-Identifier: LGPL-2.1-or-later
# SPDX-FileNotice: Part of the AddonManager.

# -*- coding: utf-8 -*-

################################################################################
## Form generated from reading UI file 'expanded_view.ui'
##
## Created by: Qt User Interface Compiler version 5.15.1
##
## WARNING! All changes made in this file will be lost when recompiling UI file!
################################################################################

from PySideWrapper import QtCore, QtWidgets


class Ui_ExpandedView(object):
    def setupUi(self, ExpandedView):
        if not ExpandedView.objectName():
            ExpandedView.setObjectName("ExpandedView")
        ExpandedView.resize(807, 141)
        sizePolicy = QtWidgets.QSizePolicy(
            QtWidgets.QSizePolicy.MinimumExpanding, QtWidgets.QSizePolicy.Preferred
        )
        sizePolicy.setHorizontalStretch(0)
        sizePolicy.setVerticalStretch(0)
        sizePolicy.setHeightForWidth(ExpandedView.sizePolicy().hasHeightForWidth())
        ExpandedView.setSizePolicy(sizePolicy)
        self.horizontalLayout_2 = QtWidgets.QHBoxLayout(ExpandedView)
        self.horizontalLayout_2.setSpacing(2)
        self.horizontalLayout_2.setObjectName("horizontalLayout_2")
        self.horizontalLayout_2.setSizeConstraint(QtWidgets.QLayout.SetNoConstraint)
        self.horizontalLayout_2.setContentsMargins(2, 0, 2, 0)
        self.labelIcon = QtWidgets.QLabel(ExpandedView)
        self.labelIcon.setObjectName("labelIcon")
        sizePolicy1 = QtWidgets.QSizePolicy(
            QtWidgets.QSizePolicy.Fixed, QtWidgets.QSizePolicy.Fixed
        )
        sizePolicy1.setHorizontalStretch(0)
        sizePolicy1.setVerticalStretch(0)
        sizePolicy1.setHeightForWidth(self.labelIcon.sizePolicy().hasHeightForWidth())
        self.labelIcon.setSizePolicy(sizePolicy1)
        self.labelIcon.setMinimumSize(QtCore.QSize(48, 48))
        self.labelIcon.setMaximumSize(QtCore.QSize(48, 48))
        self.labelIcon.setBaseSize(QtCore.QSize(48, 48))

        self.horizontalLayout_2.addWidget(self.labelIcon)

        self.horizontalSpacer = QtWidgets.QSpacerItem(
            8, 20, QtWidgets.QSizePolicy.Fixed, QtWidgets.QSizePolicy.Minimum
        )

        self.horizontalLayout_2.addItem(self.horizontalSpacer)

        self.verticalLayout = QtWidgets.QVBoxLayout()
        self.verticalLayout.setSpacing(3)
        self.verticalLayout.setObjectName("verticalLayout")
        self.horizontalLayout = QtWidgets.QHBoxLayout()
        self.horizontalLayout.setSpacing(10)
        self.horizontalLayout.setObjectName("horizontalLayout")
        self.labelPackageName = QtWidgets.QLabel(ExpandedView)
        self.labelPackageName.setObjectName("labelPackageName")

        self.horizontalLayout.addWidget(self.labelPackageName)

        self.labelVersion = QtWidgets.QLabel(ExpandedView)
        self.labelVersion.setObjectName("labelVersion")
        self.labelVersion.setTextFormat(QtCore.Qt.RichText)
        sizePolicy2 = QtWidgets.QSizePolicy(
            QtWidgets.QSizePolicy.Preferred, QtWidgets.QSizePolicy.Preferred
        )
        sizePolicy2.setHorizontalStretch(0)
        sizePolicy2.setVerticalStretch(0)
        sizePolicy2.setHeightForWidth(self.labelVersion.sizePolicy().hasHeightForWidth())
        self.labelVersion.setSizePolicy(sizePolicy2)

        self.horizontalLayout.addWidget(self.labelVersion)

        self.labelTags = QtWidgets.QLabel(ExpandedView)
        self.labelTags.setObjectName("labelTags")

        self.horizontalLayout.addWidget(self.labelTags)

        self.labelSort = QtWidgets.QLabel(ExpandedView)
        self.labelSort.setObjectName("labelSort")

        self.horizontalLayout.addWidget(self.labelSort)

        self.horizontalSpacer_2 = QtWidgets.QSpacerItem(
            40, 20, QtWidgets.QSizePolicy.Expanding, QtWidgets.QSizePolicy.Minimum
        )

        self.horizontalLayout.addItem(self.horizontalSpacer_2)

        self.verticalLayout.addLayout(self.horizontalLayout)

        self.labelDescription = QtWidgets.QLabel(ExpandedView)
        self.labelDescription.setObjectName("labelDescription")
        sizePolicy.setHeightForWidth(self.labelDescription.sizePolicy().hasHeightForWidth())
        self.labelDescription.setSizePolicy(sizePolicy)
        self.labelDescription.setTextFormat(QtCore.Qt.PlainText)
        self.labelDescription.setAlignment(
            QtCore.Qt.AlignLeading | QtCore.Qt.AlignLeft | QtCore.Qt.AlignTop
        )
        self.labelDescription.setWordWrap(True)

        self.verticalLayout.addWidget(self.labelDescription)

        self.labelMaintainer = QtWidgets.QLabel(ExpandedView)
        self.labelMaintainer.setObjectName("labelMaintainer")
        sizePolicy2.setHeightForWidth(self.labelMaintainer.sizePolicy().hasHeightForWidth())
        self.labelMaintainer.setSizePolicy(sizePolicy2)
        self.labelMaintainer.setAlignment(
            QtCore.Qt.AlignLeading | QtCore.Qt.AlignLeft | QtCore.Qt.AlignTop
        )
        self.labelMaintainer.setWordWrap(False)

        self.verticalLayout.addWidget(self.labelMaintainer)

        self.horizontalLayout_2.addLayout(self.verticalLayout)

        self.labelStatus = QtWidgets.QLabel(ExpandedView)
        self.labelStatus.setObjectName("labelStatus")
        self.labelStatus.setTextFormat(QtCore.Qt.RichText)
        self.labelStatus.setAlignment(
            QtCore.Qt.AlignRight | QtCore.Qt.AlignTrailing | QtCore.Qt.AlignVCenter
        )

        self.horizontalLayout_2.addWidget(self.labelStatus)

        self.retranslateUi(ExpandedView)

        QtCore.QMetaObject.connectSlotsByName(ExpandedView)

    # setupUi

    def retranslateUi(self, ExpandedView):
        #        ExpandedView.setWindowTitle(QCoreApplication.translate("ExpandedView", "Form", None))
        self.labelIcon.setText(QtCore.QCoreApplication.translate("ExpandedView", "Icon", None))
        self.labelPackageName.setText(
            QtCore.QCoreApplication.translate("ExpandedView", "<h1>Package Name</h1>", None)
        )
        self.labelVersion.setText(
            QtCore.QCoreApplication.translate("ExpandedView", "Version", None)
        )
        self.labelTags.setText(QtCore.QCoreApplication.translate("ExpandedView", "(tags)", None))
        self.labelDescription.setText(
            QtCore.QCoreApplication.translate("ExpandedView", "Description", None)
        )
        self.labelMaintainer.setText(
            QtCore.QCoreApplication.translate("ExpandedView", "Maintainer", None)
        )
        self.labelStatus.setText(
            QtCore.QCoreApplication.translate("ExpandedView", "Update available", None)
        )

    # retranslateUi



# SPDX-License-Identifier: LGPL-2.1-or-later

# ***************************************************************************
# *   Copyright (c) 2020 sliptonic <shopinghewoods@gmail.com>               *
# *                                                                         *
# *   This file is part of the FreeCAD CAx development system.              *
# *                                                                         *
# *   This program is free software; you can redistribute it and/or modify  *
# *   it under the terms of the GNU Lesser General Public License (LGPL)    *
# *   as published by the Free Software Foundation; either version 2 of     *
# *   the License, or (at your option) any later version.                   *
# *   for detail see the LICENCE text file.                                 *
# *                                                                         *
# *   FreeCAD is distributed in the hope that it will be useful,            *
# *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
# *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
# *   GNU Lesser General Public License for more details.                   *
# *                                                                         *
# *   You should have received a copy of the GNU Library General Public     *
# *   License along with FreeCAD; if not, write to the Free Software        *
# *   Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  *
# *   USA                                                                   *
# *                                                                         *
# ***************************************************************************/

import FreeCAD
import Part
import Path
import PathScripts.PathUtils as PathUtils
import datetime
import importDXF
from builtins import open as pyopen

TOOLTIP = """
This is a postprocessor file for the Path workbench. It is used to
take a pseudo-G-code fragment outputted by a Path object, and output
a dxf file.
Operations are output to layers.
vertical moves are ignore
All path moves are flattened to z=0

Does NOT remove redundant lines.  If you have multiple step-downs in your
operation, you'll get multiple redundant lines in your dxf.

import dxf_post
"""

TOOLTIP_ARGS = """
Arguments for dxf:
"""
now = datetime.datetime.now()

# # These globals set common customization preferences
OUTPUT_HEADER = True

if False:
    Path.Log.setLevel(Path.Log.Level.DEBUG, Path.Log.thisModule())
    Path.Log.trackModule(Path.Log.thisModule())
else:
    Path.Log.setLevel(Path.Log.Level.INFO, Path.Log.thisModule())
Path.Log.setLevel(Path.Log.Level.INFO, Path.Log.thisModule())


def processArguments(argstring):
    pass
    # global OUTPUT_HEADER


def export(objectslist, filename, argstring):
    doc = FreeCAD.ActiveDocument
    print("postprocessing...")
    layers = []
    processArguments(argstring)
    for i in objectslist:
        result = parse(i)
        if len(result) > 0:
            layername = i.Name
            grp = doc.addObject("App::DocumentObjectGroup", layername)
            for o in result:
                o.adjustRelativeLinks(grp)
                grp.addObject(o)
            layers.append(grp)

    dxfWrite(layers, filename)


def dxfWrite(objlist, filename):
    importDXF.export(objlist, filename)


def parse(pathobj):
    """accepts a Path object.  Returns a list of wires"""

    feedcommands = Path.Geom.CmdMove
    rapidcommands = Path.Geom.CmdMoveRapid

    edges = []
    objlist = []

    # Gotta start somewhere.  Assume 0,0,0
    curPoint = FreeCAD.Vector(0, 0, 0)
    for c in PathUtils.getPathWithPlacement(pathobj).Commands:
        Path.Log.debug("{} -> {}".format(curPoint, c))
        if "Z" in c.Parameters:
            newparams = c.Parameters
            newparams.pop("Z", None)
            flatcommand = Path.Command(c.Name, newparams)
            c.Parameters = newparams
        else:
            flatcommand = c

        # ignore gcode that isn't moving
        if flatcommand.Name not in feedcommands + rapidcommands:
            Path.Log.debug("non move")
            continue

        # ignore pure vertical feed and rapid
        if (
            flatcommand.Parameters.get("X", curPoint.x) == curPoint.x
            and flatcommand.Parameters.get("Y", curPoint.y) == curPoint.y
        ):
            Path.Log.debug("vertical")
            continue

        # feeding move.  Build an edge
        if flatcommand.Name in feedcommands:
            edges.append(Path.Geom.edgeForCmd(flatcommand, curPoint))
            Path.Log.debug("feeding move")

        # update the curpoint
        curPoint.x = flatcommand.Parameters.get("X", curPoint.x)
        curPoint.y = flatcommand.Parameters.get("Y", curPoint.y)

    if len(edges) > 0:
        candidates = Part.sortEdges(edges)
        for c in candidates:
            obj = FreeCAD.ActiveDocument.addObject("Part::Feature", "Wire")
            obj.Shape = Part.Wire(c)
            objlist.append(obj)

    return objlist



# SPDX-License-Identifier: LGPL-2.1-or-later

# ***************************************************************************
# *   Copyright (c) 2014 sliptonic <shopinthewoods@gmail.com>               *
# *                                                                         *
# *   Reabased changes from relative_post.py to linuxcnc_post.py            *
# *   and updated functionality for old EDM machines by alromh87            *
# *                                                                         *
# *   This file is part of the FreeCAD CAx development system.              *
# *                                                                         *
# *   This program is free software; you can redistribute it and/or modify  *
# *   it under the terms of the GNU Lesser General Public License (LGPL)    *
# *   as published by the Free Software Foundation; either version 2 of     *
# *   the License, or (at your option) any later version.                   *
# *   for detail see the LICENCE text file.                                 *
# *                                                                         *
# *   FreeCAD is distributed in the hope that it will be useful,            *
# *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
# *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
# *   GNU Lesser General Public License for more details.                   *
# *                                                                         *
# *   You should have received a copy of the GNU Library General Public     *
# *   License along with FreeCAD; if not, write to the Free Software        *
# *   Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  *
# *   USA                                                                   *
# *                                                                         *
# ***************************************************************************

import FreeCAD
from FreeCAD import Units
import Path
import argparse
import datetime
import shlex
import Path.Base.Util as PathUtil
import Path.Post.Utils as PostUtils
import PathScripts.PathUtils as PathUtils
from builtins import open as pyopen

TOOLTIP = """
This is a postprocessor file for the Path workbench. It is used to
take a pseudo-G-code fragment outputted by a Path object, and output
real G-code suitable for a Wire EDM CNC. This postprocessor, once placed
in the appropriate PathScripts folder, can be used directly from inside
FreeCAD, via the GUI importer or via python scripts with:

import wedm_post
wedm_post.export(object,"/path/to/file.ncc","")
"""

now = datetime.datetime.now()

parser = argparse.ArgumentParser(prog="wedm", add_help=False)
parser.add_argument("--no-header", action="store_true", help="suppress header output")
parser.add_argument("--no-comments", action="store_true", help="suppress comment output")
parser.add_argument(
    "--comments-character", default="", help="Use provided character before comments"
)
parser.add_argument(
    "--command-space", default=" ", help="Use provided character as space in commands"
)
parser.add_argument("--endline-character", default="", help="Use provided character at end of line")
parser.add_argument("--line-numbers", action="store_true", help="prefix with line numbers")
parser.add_argument(
    "--no-show-editor",
    action="store_true",
    help="don't pop up editor before writing output",
)
parser.add_argument("--scale", default="1", help="Scale factor for coordinates")
parser.add_argument("--precision", default="3", help="number of digits of precision, default=3")
parser.add_argument("--fixed-length", default="0", help="use fixed length coordinates, default=0")
parser.add_argument(
    "--preamble",
    help='set commands to be issued before the first command, default="G17 G54 G40 G49 G80 G90\\n"',
)
parser.add_argument(
    "--postamble",
    help='set commands to be issued after the last command, default="M05\\nG17 G54 G90 G80 G40\\nM2\\n"',
)
parser.add_argument(
    "--inches", action="store_true", help="Convert output for US imperial mode (G20)"
)
parser.add_argument(
    "--modal",
    action="store_true",
    help="Output the Same G-command Name USE NonModal Mode",
)
parser.add_argument("--axis-modal", action="store_true", help="Output the Same Axis Value Mode")
parser.add_argument(
    "--no-tlo",
    action="store_true",
    help="suppress tool length offset (G43) following tool changes",
)
parser.add_argument(
    "--use-rapids",
    action="store_true",
    help="Allow G0 on output",
)
parser.add_argument(
    "--omit-units",
    action="store_true",
    help="Don't output units G20/G21",
)
parser.add_argument("--relative", action="store_true", help="Generate Relative GCODE")
parser.add_argument(
    "--two-digit-codes",
    action="store_true",
    help="Add trailing 0 to codes lower than 10 (G1 -> G01)",
)
parser.add_argument("--force-sign", action="store_true", help="Always add sign to coordinates")
parser.add_argument(
    "--ignore-operations",
    default="",
    help="Ignore provided operations, use Labels and separate with ','",
)

TOOLTIP_ARGS = parser.format_help()

# These globals set common customization preferences
OUTPUT_COMMENTS = True
COMMENT_CHAR = ""
COMMAND_SPACE = " "
ENDLINE = ""
OUTPUT_HEADER = True
OUTPUT_LINE_NUMBERS = False
SHOW_EDITOR = True
MODAL = False  # if true commands are suppressed if the same as previous line.
USE_TLO = True  # if true G43 will be output following tool changes
OUTPUT_DOUBLES = True  # if false duplicate axis values are suppressed if the same as previous line.
LINENR = 100  # line number starting value
USE_RAPIDS = False
OMIT_UNITS = False
RELATIVE_GCODE = False
TWO_DIGIT_CODES = False
FORCE_SIGN = False

# These globals will be reflected in the Machine configuration of the project
UNITS = "G21"  # G21 for metric, G20 for us standard
UNIT_SPEED_FORMAT = "mm/min"
UNIT_FORMAT = "mm"

MACHINE_NAME = "Wire EDM"
CORNER_MIN = {"x": 0, "y": 0, "z": 0}
CORNER_MAX = {"x": 500, "y": 300, "z": 300}
PRECISION = 3

# Preamble text will appear at the beginning of the GCODE output file.
PREAMBLE = """G17 G54 G40 G49 G80 G90
"""


# Postamble text will appear following the last operation.
POSTAMBLE = """M05
G17 G54 G90 G80 G40
M2
"""

# Pre operation text will be inserted before every operation
PRE_OPERATION = """"""

# Post operation text will be inserted after every operation
POST_OPERATION = """"""

# Tool Change commands will be inserted before a tool change
TOOL_CHANGE = """"""


def processArguments(argstring):
    global OUTPUT_HEADER
    global OUTPUT_COMMENTS
    global COMMENT_CHAR
    global COMMAND_SPACE
    global ENDLINE
    global OUTPUT_LINE_NUMBERS
    global SHOW_EDITOR
    global SCALE
    global PRECISION
    global FIXED_LENGTH
    global PREAMBLE
    global POSTAMBLE
    global UNITS
    global UNIT_SPEED_FORMAT
    global UNIT_FORMAT
    global MODAL
    global USE_TLO
    global OUTPUT_DOUBLES
    global LOCATION  # keep track for incremental
    global USE_RAPIDS
    global OMIT_UNITS
    global RELATIVE_GCODE
    global TWO_DIGIT_CODES
    global FORCE_SIGN
    global IGNORE_OPERATIONS

    try:
        args = parser.parse_args(shlex.split(argstring))
        if args.no_header:
            OUTPUT_HEADER = False
        if args.no_comments:
            OUTPUT_COMMENTS = False
        COMMENT_CHAR = args.comments_character
        COMMAND_SPACE = args.command_space
        ENDLINE = args.endline_character
        if args.line_numbers:
            OUTPUT_LINE_NUMBERS = True
        if args.no_show_editor:
            SHOW_EDITOR = False
        SCALE = int(args.scale)
        PRECISION = args.precision
        FIXED_LENGTH = int(args.fixed_length)
        if args.preamble is not None:
            PREAMBLE = args.preamble.replace("\\n", "\n")
        if args.postamble is not None:
            POSTAMBLE = args.postamble.replace("\\n", "\n")
        if args.inches:
            UNITS = "G20"
            UNIT_SPEED_FORMAT = "in/min"
            UNIT_FORMAT = "in"
            PRECISION = 4
        if args.modal:
            MODAL = True
        if args.no_tlo:
            USE_TLO = False
        if args.axis_modal:
            OUTPUT_DOUBLES = False
        if args.use_rapids:
            USE_RAPIDS = True
        if args.omit_units:
            OMIT_UNITS = True
        if args.relative:
            print("relative")
            RELATIVE_GCODE = True
            # PREAMBLE = PREAMBLE.replace("G90", "")
            # PREAMBLE = PREAMBLE.replace("G91", "")
            PREAMBLE += "\nG91"
            LOCATION = {"X": 0, "Y": 0, "Z": 0}
        if args.two_digit_codes:
            TWO_DIGIT_CODES = True
        if args.force_sign:
            FORCE_SIGN = True
        IGNORE_OPERATIONS = args.ignore_operations.split(",")

    except Exception as e:
        print(e)
        return False

    return True


def export(objectslist, filename, argstring):
    if not processArguments(argstring):
        return None
    global UNITS
    global UNIT_FORMAT
    global UNIT_SPEED_FORMAT
    global MACHINE_NAME

    for obj in objectslist:
        if not hasattr(obj, "Path"):
            print(
                "the object " + obj.Name + " is not a path. Please select only path and Compounds."
            )
            return None

    print("postprocessing...")
    gcode = ""

    # write header
    if OUTPUT_HEADER:
        gcode += linenumber() + COMMENT_CHAR + "(Exported by FreeCAD)\n"
        gcode += linenumber() + COMMENT_CHAR + "(Post Processor: " + __name__ + ")\n"
        gcode += linenumber() + COMMENT_CHAR + "(Output Time:" + str(now) + ")\n"

    # Write the preamble
    if OUTPUT_COMMENTS:
        gcode += linenumber() + COMMENT_CHAR + "(begin preamble)\n"
    for line in PREAMBLE.splitlines():
        gcode += linenumber() + line + "\n"
    if not OMIT_UNITS:
        gcode += linenumber() + UNITS + ENDLINE + "\n"

    for obj in objectslist:

        # Skip inactive operations
        if not PathUtil.activeForOp(obj):
            continue

        # fetch machine details
        job = PathUtils.findParentJob(obj)

        # TODO: Not sure of the use
        if hasattr(job, "MachineName"):
            MACHINE_NAME = job.MachineName

        if hasattr(job, "MachineUnits"):
            print(job.MachineUnits)
            if job.MachineUnits == "Metric":
                UNITS = "G21"
                UNIT_FORMAT = "mm"
                UNIT_SPEED_FORMAT = "mm/min"
            else:
                UNITS = "G20"
                UNIT_FORMAT = "in"
                UNIT_SPEED_FORMAT = "in/min"

        # ignore selected operations
        if obj.Label in IGNORE_OPERATIONS:
            continue
        # do the pre_op
        if OUTPUT_COMMENTS:
            gcode += linenumber() + COMMENT_CHAR + "(begin operation: %s)\n" % obj.Label
            gcode += (
                linenumber()
                + COMMENT_CHAR
                + "(machine: %s, %s)\n" % (MACHINE_NAME, UNIT_SPEED_FORMAT)
            )
        for line in PRE_OPERATION.splitlines(True):
            gcode += linenumber() + line

        # get coolant mode
        coolantMode = PathUtil.coolantModeForOp(obj)

        # turn coolant on if required
        if OUTPUT_COMMENTS:
            if not coolantMode == "None":
                gcode += linenumber() + COMMENT_CHAR + "(Coolant On:" + coolantMode + ")\n"
        if coolantMode == "Flood":
            gcode += linenumber() + "M8" + "\n"
        if coolantMode == "Mist":
            gcode += linenumber() + "M7" + "\n"

        # process the operation gcode
        gcode += parse(obj)

        # do the post_op
        if OUTPUT_COMMENTS:
            gcode += linenumber() + COMMENT_CHAR + "(finish operation: %s)\n" % obj.Label
        for line in POST_OPERATION.splitlines(True):
            gcode += linenumber() + line

        # turn coolant off if required
        if not coolantMode == "None":
            if OUTPUT_COMMENTS:
                gcode += linenumber() + COMMENT_CHAR + "(Coolant Off:" + coolantMode + ")\n"
            gcode += linenumber() + "M9" + "\n"

    # do the post_amble
    if OUTPUT_COMMENTS:
        gcode += linenumber() + COMMENT_CHAR + "(begin postamble)\n"
    for line in POSTAMBLE.splitlines():
        gcode += linenumber() + line + "\n"

    if FreeCAD.GuiUp and SHOW_EDITOR:
        final = gcode
        if len(gcode) > 100000:
            print("Skipping editor since output is greater than 100kb")
        else:
            dia = PostUtils.GCodeEditorDialog()
            dia.editor.setText(gcode)
            result = dia.exec_()
            if result:
                final = dia.editor.toPlainText()
    else:
        final = gcode

    print("done postprocessing.")

    if not filename == "-":
        gfile = pyopen(filename, "w")
        gfile.write(final)
        gfile.close()

    return final


def linenumber():
    global LINENR
    if OUTPUT_LINE_NUMBERS is True:
        LINENR += 10
        return "N" + str(LINENR) + COMMAND_SPACE
    return ""


def parse(pathobj):
    global PRECISION
    global MODAL
    global OUTPUT_DOUBLES
    global UNIT_FORMAT
    global UNIT_SPEED_FORMAT
    global RELATIVE_GCODE
    global LOCATION

    out = ""
    lastcommand = None
    precision_string = "." + str(PRECISION) + "f"
    currLocation = {}  # keep track for incremental

    # the order of parameters
    # linuxcnc doesn't want K properties on XY plane  Arcs need work.
    params = [
        "X",
        "Y",
        "Z",
        "A",
        "B",
        "C",
        "I",
        "J",
        "F",
        "S",
        "T",
        "Q",
        "R",
        "L",
        "H",
        "D",
        "P",
    ]
    firstmove = Path.Command("G0", {"X": 0, "Y": 0, "Z": 0, "F": 0.0})
    currLocation.update(firstmove.Parameters)  # set First location Parameters

    if hasattr(pathobj, "Group"):  # We have a compound or project.
        # if OUTPUT_COMMENTS:
        #     out += linenumber() + COMMENT_CHAR + "(compound: " + pathobj.Label + ")\n"
        for p in pathobj.Group:
            out += parse(p)
        return out
    else:  # parsing simple path

        # groups might contain non-path things like stock.
        if not hasattr(pathobj, "Path"):
            return out

        # if OUTPUT_COMMENTS:
        #     out += linenumber() + COMMENT_CHAR + "(" + pathobj.Label + ")\n"

        for c in PathUtils.getPathWithPlacement(pathobj).Commands:
            # For Debug Only
            #     if OUTPUT_COMMENTS:
            #         out += linenumber() + COMMENT_CHAR + "(" + str(c) + ")\n"

            outstring = []
            command = c.Name
            if not USE_RAPIDS and (command == "G0" or command == "G00"):
                command = "G1"
            outstring.append(command)

            # if modal: suppress the command if it is the same as the last one
            if MODAL is True:
                if command == lastcommand:
                    outstring.pop(0)

            if command.startswith("("):  # command is a comment
                if OUTPUT_COMMENTS:  # Edit comment  with COMMENT_CHAR
                    outstring.insert(0, COMMENT_CHAR)
                else:
                    continue

            move_comands = ["G0", "G00", "G1", "G01"]
            # Now add the remaining parameters in order
            for param in params:
                if param in c.Parameters:
                    if param == "F" and (
                        currLocation[param] != c.Parameters[param] or OUTPUT_DOUBLES
                    ):
                        if command not in [
                            "G0",
                            "G00",
                        ]:  # linuxcnc doesn't use rapid speeds
                            speed = Units.Quantity(c.Parameters["F"], FreeCAD.Units.Velocity)
                            if speed.getValueAs(UNIT_SPEED_FORMAT) > 0.0:
                                outstring.append(
                                    param
                                    + format(
                                        float(speed.getValueAs(UNIT_SPEED_FORMAT)),
                                        precision_string,
                                    )
                                )
                        else:
                            continue
                    elif param == "T":
                        outstring.append(param + str(int(c.Parameters["T"])))
                    elif param == "H":
                        outstring.append(param + str(int(c.Parameters["H"])))
                    elif param == "D":
                        outstring.append(param + str(int(c.Parameters["D"])))
                    elif param == "S":
                        outstring.append(param + str(int(c.Parameters["S"])))
                    # For wire EDM we ignore values for Z
                    elif param == "Z":
                        continue
                    else:
                        if (
                            (not OUTPUT_DOUBLES)
                            and (param in currLocation)
                            and (currLocation[param] == c.Parameters[param])
                        ):
                            continue
                        else:
                            if RELATIVE_GCODE and (param != "I" and param != "J"):
                                pos = Units.Quantity(
                                    c.Parameters[param] - currLocation[param], FreeCAD.Units.Length
                                )
                                print(
                                    f"currlocation: {currLocation[param]} param: {c.Parameters[param]} pos: {pos}"
                                )
                                if pos == 0:
                                    # Remove no movement
                                    continue
                            else:
                                pos = Units.Quantity(c.Parameters[param], FreeCAD.Units.Length)

                            pos = pos * SCALE
                            sign = ""
                            if pos >= 0 and FORCE_SIGN:
                                sign = "+"

                            stringout = sign + format(
                                float(pos.getValueAs(UNIT_FORMAT)), precision_string
                            )

                            # Remove unneeded 0s on incremental moves this is needed since some numbers are zero only after applying precision
                            if (
                                RELATIVE_GCODE
                                and command in move_comands
                                and (float(stringout) == 0)
                            ):
                                continue
                            # Force trailing zeros
                            if FIXED_LENGTH > 0:
                                extra = 1 if stringout[0] in ["+", "-"] else 0
                                stringout = stringout.zfill(FIXED_LENGTH + extra)
                            outstring.append(param + stringout)

            currLocation.update(c.Parameters)

            # Check for Tool Change:
            if command == "M6":
                # stop the spindle
                out += linenumber() + "M5\n"
                for line in TOOL_CHANGE.splitlines(True):
                    out += linenumber() + line

                # add height offset
                if USE_TLO:
                    tool_height = "\nG43 H" + str(int(c.Parameters["T"]))
                    outstring.append(tool_height)

            if command == "message":
                if OUTPUT_COMMENTS is False:
                    out = []
                else:
                    outstring.pop(0)  # remove the command

            # prepend a line number and append a newline
            if len(outstring) >= 1:
                if len(outstring) == 1 and (outstring[0] == "G0" or outstring[0] == "G1"):
                    # Don't write empty moves (Generated when Z moves are ignored for EDM)
                    print("Ignoring: " + command + ":" + outstring[0])
                    continue
                if TWO_DIGIT_CODES and len(command) == 2 and outstring[0] == command:
                    outstring[0] = command[0] + "0" + command[1]
                if OUTPUT_LINE_NUMBERS:
                    outstring.insert(0, (linenumber()))

                # append the line to the final output
                start = True
                for w in outstring:
                    if start:
                        start = False
                    else:
                        out += COMMAND_SPACE
                    out += w
                # Note: Do *not* strip `out`, since that forces the allocation
                # of a contiguous string & thus quadratic complexity.
                out += ENDLINE + "\n"

            # store the latest command after written
            lastcommand = command

        return out


# print(__name__ + " gcode postprocessor loaded.")



# ***************************************************************************
# *   Copyright (c) 2019 Bernd Hahnebach <bernd@bimstatik.org>              *
# *   Copyright (c) 2020 Sudhanshu Dubey <sudhanshu.thethunder@gmail.com    *
# *                                                                         *
# *   This file is part of the FreeCAD CAx development system.              *
# *                                                                         *
# *   This program is free software; you can redistribute it and/or modify  *
# *   it under the terms of the GNU Lesser General Public License (LGPL)    *
# *   as published by the Free Software Foundation; either version 2 of     *
# *   the License, or (at your option) any later version.                   *
# *   for detail see the LICENCE text file.                                 *
# *                                                                         *
# *   This program is distributed in the hope that it will be useful,       *
# *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
# *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
# *   GNU Library General Public License for more details.                  *
# *                                                                         *
# *   You should have received a copy of the GNU Library General Public     *
# *   License along with this program; if not, write to the Free Software   *
# *   Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  *
# *   USA                                                                   *
# *                                                                         *
# ***************************************************************************

import FreeCAD

import Fem
import ObjectsFem

from .manager import get_meshname
from .manager import init_doc
from .meshes import generate_mesh


def setup_boxanalysisbase(doc=None, solvertype="ccxtools", test_mode=False):

    # init FreeCAD document
    if doc is None:
        doc = init_doc()

    # geometric objects
    # object name is important in this base setup method
    # all module which use this base setup, use the object name to find the object
    geom_obj = doc.addObject("Part::Box", "Box")
    geom_obj.Height = geom_obj.Width = geom_obj.Length = 10
    doc.recompute()

    if FreeCAD.GuiUp:
        geom_obj.ViewObject.Document.activeView().viewAxonometric()
        geom_obj.ViewObject.Document.activeView().fitAll()

    # analysis
    analysis = ObjectsFem.makeAnalysis(doc, "Analysis")

    # material
    material_obj = ObjectsFem.makeMaterialSolid(doc, "MechanicalMaterial")
    mat = material_obj.Material
    mat["Name"] = "Steel-Generic"
    mat["YoungsModulus"] = "200000 MPa"
    mat["PoissonRatio"] = "0.30"
    mat["Density"] = "7900 kg/m^3"
    material_obj.Material = mat
    analysis.addObject(material_obj)

    # mesh
    femmesh_obj = analysis.addObject(ObjectsFem.makeMeshGmsh(doc, get_meshname()))[0]
    femmesh_obj.Shape = geom_obj
    femmesh_obj.SecondOrderLinear = False
    femmesh_obj.CharacteristicLengthMin = "8.0 mm"
    femmesh_obj.ElementOrder = "2nd"

    # generate the mesh
    success = False
    if not test_mode:
        success = generate_mesh.mesh_from_mesher(femmesh_obj, "gmsh")
    if not success:
        # try to create from existing rough mesh
        from .meshes.mesh_boxanalysis_tetra10 import create_nodes, create_elements

        fem_mesh = generate_mesh.mesh_from_existing(create_nodes, create_elements)
        femmesh_obj.FemMesh = fem_mesh

    doc.recompute()
    return doc



# SPDX-License-Identifier: LGPL-2.1-or-later

# ***************************************************************************
# *   Copyright (c) 2020 sliptonic <shopinthewoods@gmail.com>               *
# *   Copyright (c) 2020 russ4262 <russ4262@gmail.com>                      *
# *                                                                         *
# *   This program is free software; you can redistribute it and/or modify  *
# *   it under the terms of the GNU Lesser General Public License (LGPL)    *
# *   as published by the Free Software Foundation; either version 2 of     *
# *   the License, or (at your option) any later version.                   *
# *   for detail see the LICENCE text file.                                 *
# *                                                                         *
# *   This program is distributed in the hope that it will be useful,       *
# *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
# *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
# *   GNU Library General Public License for more details.                  *
# *                                                                         *
# *   You should have received a copy of the GNU Library General Public     *
# *   License along with this program; if not, write to the Free Software   *
# *   Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  *
# *   USA                                                                   *
# *                                                                         *
# ***************************************************************************

from PySide import QtCore
from PySide.QtCore import QT_TRANSLATE_NOOP
import FreeCAD
import FreeCADGui
import Path
import Path.Base.Gui.Util as PathGuiUtil
import Path.Op.Gui.Base as PathOpGui
import Path.Op.Waterline as PathWaterline

__title__ = "CAM Waterline Operation UI"
__author__ = "sliptonic (Brad Collette), russ4262 (Russell Johnson)"
__url__ = "https://www.freecad.org"
__doc__ = "Waterline operation page controller and command implementation."

translate = FreeCAD.Qt.translate

if False:
    Path.Log.setLevel(Path.Log.Level.DEBUG, Path.Log.thisModule())
    Path.Log.trackModule(Path.Log.thisModule())
else:
    Path.Log.setLevel(Path.Log.Level.INFO, Path.Log.thisModule())


class TaskPanelOpPage(PathOpGui.TaskPanelPage):
    """Page controller class for the Waterline operation."""

    def initPage(self, obj):
        self.setTitle("Waterline - " + obj.Label)
        self.updateVisibility()

    def getForm(self):
        """getForm() ... returns UI"""
        form = FreeCADGui.PySideUic.loadUi(":/panels/PageOpWaterlineEdit.ui")
        comboToPropertyMap = [
            ("algorithmSelect", "Algorithm"),
            ("boundBoxSelect", "BoundBox"),
            ("layerMode", "LayerMode"),
            ("cutPattern", "CutPattern"),
        ]
        enumTups = PathWaterline.ObjectWaterline.propertyEnumerations(dataType="raw")
        PathGuiUtil.populateCombobox(form, enumTups, comboToPropertyMap)
        return form

    def getFields(self, obj):
        """getFields(obj) ... transfers values from UI to obj's properties"""
        self.updateToolController(obj, self.form.toolController)
        self.updateCoolant(obj, self.form.coolantController)

        if obj.Algorithm != str(self.form.algorithmSelect.currentData()):
            obj.Algorithm = str(self.form.algorithmSelect.currentData())

        if obj.BoundBox != str(self.form.boundBoxSelect.currentData()):
            obj.BoundBox = str(self.form.boundBoxSelect.currentData())

        if obj.LayerMode != str(self.form.layerMode.currentData()):
            obj.LayerMode = str(self.form.layerMode.currentData())

        if obj.CutPattern != str(self.form.cutPattern.currentData()):
            obj.CutPattern = str(self.form.cutPattern.currentData())

        PathGuiUtil.updateInputField(obj, "BoundaryAdjustment", self.form.boundaryAdjustment)

        if obj.StepOver != self.form.stepOver.value():
            obj.StepOver = self.form.stepOver.value()

        PathGuiUtil.updateInputField(obj, "SampleInterval", self.form.sampleInterval)

        if obj.OptimizeLinearPaths != self.form.optimizeEnabled.isChecked():
            obj.OptimizeLinearPaths = self.form.optimizeEnabled.isChecked()

    def setFields(self, obj):
        """setFields(obj) ... transfers obj's property values to UI"""
        self.setupToolController(obj, self.form.toolController)
        self.setupCoolant(obj, self.form.coolantController)
        self.selectInComboBox(obj.Algorithm, self.form.algorithmSelect)
        self.selectInComboBox(obj.BoundBox, self.form.boundBoxSelect)
        self.selectInComboBox(obj.LayerMode, self.form.layerMode)
        self.selectInComboBox(obj.CutPattern, self.form.cutPattern)
        self.form.boundaryAdjustment.setText(
            FreeCAD.Units.Quantity(obj.BoundaryAdjustment.Value, FreeCAD.Units.Length).UserString
        )
        self.form.stepOver.setValue(obj.StepOver)
        self.form.sampleInterval.setText(
            FreeCAD.Units.Quantity(obj.SampleInterval.Value, FreeCAD.Units.Length).UserString
        )

        if obj.OptimizeLinearPaths:
            self.form.optimizeEnabled.setCheckState(QtCore.Qt.Checked)
        else:
            self.form.optimizeEnabled.setCheckState(QtCore.Qt.Unchecked)

        self.updateVisibility()

    def getSignalsForUpdate(self, obj):
        """getSignalsForUpdate(obj) ... return list of signals for updating obj"""
        signals = []
        signals.append(self.form.toolController.currentIndexChanged)
        signals.append(self.form.coolantController.currentIndexChanged)
        signals.append(self.form.algorithmSelect.currentIndexChanged)
        signals.append(self.form.boundBoxSelect.currentIndexChanged)
        signals.append(self.form.layerMode.currentIndexChanged)
        signals.append(self.form.cutPattern.currentIndexChanged)
        signals.append(self.form.boundaryAdjustment.editingFinished)
        signals.append(self.form.stepOver.editingFinished)
        signals.append(self.form.sampleInterval.editingFinished)
        if hasattr(self.form.optimizeEnabled, "checkStateChanged"):  # Qt version >= 6.7.0
            signals.append(self.form.optimizeEnabled.checkStateChanged)
        else:  # Qt version < 6.7.0
            signals.append(self.form.optimizeEnabled.stateChanged)

        return signals

    def updateVisibility(self, sentObj=None):
        """updateVisibility(sentObj=None)... Updates visibility of Tasks panel objects."""
        Algorithm = self.form.algorithmSelect.currentData()
        self.form.optimizeEnabled.hide()  # Has no independent QLabel object

        if Algorithm == "OCL Dropcutter":
            self.form.cutPattern.hide()
            self.form.cutPattern_label.hide()
            self.form.boundaryAdjustment.hide()
            self.form.boundaryAdjustment_label.hide()
            self.form.stepOver.hide()
            self.form.stepOver_label.hide()
            self.form.sampleInterval.show()
            self.form.sampleInterval_label.show()
        elif Algorithm == "Experimental":
            self.form.cutPattern.show()
            self.form.boundaryAdjustment.show()
            self.form.cutPattern_label.show()
            self.form.boundaryAdjustment_label.show()
            if self.form.cutPattern.currentData() == "None":
                self.form.stepOver.hide()
                self.form.stepOver_label.hide()
            else:
                self.form.stepOver.show()
                self.form.stepOver_label.show()
            self.form.sampleInterval.hide()
            self.form.sampleInterval_label.hide()

    def registerSignalHandlers(self, obj):
        self.form.algorithmSelect.currentIndexChanged.connect(self.updateVisibility)
        self.form.cutPattern.currentIndexChanged.connect(self.updateVisibility)


Command = PathOpGui.SetupOperation(
    "Waterline",
    PathWaterline.Create,
    TaskPanelOpPage,
    "CAM_Waterline",
    QT_TRANSLATE_NOOP("CAM_Waterline", "Waterline"),
    QT_TRANSLATE_NOOP("CAM_Waterline", "Create a Waterline toolpath from a model"),
    PathWaterline.SetupProperties,
)

FreeCAD.Console.PrintLog("Loading PathWaterlineGui... done\n")



# SPDX-License-Identifier: LGPL-2.1-or-later

# ***************************************************************************
# *                                                                         *
# *   Copyright (c) 2024 Yorik van Havre <yorik@uncreated.net>              *
# *                                                                         *
# *   This file is part of FreeCAD.                                         *
# *                                                                         *
# *   FreeCAD is free software: you can redistribute it and/or modify it    *
# *   under the terms of the GNU Lesser General Public License as           *
# *   published by the Free Software Foundation, either version 2.1 of the  *
# *   License, or (at your option) any later version.                       *
# *                                                                         *
# *   FreeCAD is distributed in the hope that it will be useful, but        *
# *   WITHOUT ANY WARRANTY; without even the implied warranty of            *
# *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU      *
# *   Lesser General Public License for more details.                       *
# *                                                                         *
# *   You should have received a copy of the GNU Lesser General Public      *
# *   License along with FreeCAD. If not, see                               *
# *   <https://www.gnu.org/licenses/>.                                      *
# *                                                                         *
# ***************************************************************************

"""BIM Panel-related Arch_"""

import FreeCAD
import FreeCADGui

QT_TRANSLATE_NOOP = FreeCAD.Qt.QT_TRANSLATE_NOOP
translate = FreeCAD.Qt.translate

PARAMS = FreeCAD.ParamGet("User parameter:BaseApp/Preferences/Mod/BIM")


class Arch_Pipe:
    "the Arch Pipe command definition"

    def GetResources(self):

        return {
            "Pixmap": "Arch_Pipe",
            "MenuText": QT_TRANSLATE_NOOP("Arch_Pipe", "Pipe"),
            "Accel": "P, I",
            "ToolTip": QT_TRANSLATE_NOOP(
                "Arch_Pipe", "Creates a pipe object from a given wire or line"
            ),
        }

    def IsActive(self):

        v = hasattr(FreeCADGui.getMainWindow().getActiveWindow(), "getSceneGraph")
        return v

    def Activated(self):

        s = FreeCADGui.Selection.getSelection()
        if s:
            for obj in s:
                if hasattr(obj, "Shape"):
                    if len(obj.Shape.Wires) == 1:
                        FreeCAD.ActiveDocument.openTransaction(translate("Arch", "Create Pipe"))
                        FreeCADGui.addModule("Arch")
                        FreeCADGui.addModule("Draft")
                        FreeCADGui.doCommand(
                            "obj = Arch.makePipe(FreeCAD.ActiveDocument." + obj.Name + ")"
                        )
                        FreeCADGui.doCommand("Draft.autogroup(obj)")
                        FreeCAD.ActiveDocument.commitTransaction()
        else:
            FreeCAD.ActiveDocument.openTransaction(translate("Arch", "Create Pipe"))
            FreeCADGui.addModule("Arch")
            FreeCADGui.addModule("Draft")
            FreeCADGui.doCommand("obj = Arch.makePipe()")
            FreeCADGui.doCommand("Draft.autogroup(obj)")
            FreeCAD.ActiveDocument.commitTransaction()
        FreeCAD.ActiveDocument.recompute()


class Arch_PipeConnector:
    "the Arch Pipe command definition"

    def GetResources(self):

        return {
            "Pixmap": "Arch_PipeConnector",
            "MenuText": QT_TRANSLATE_NOOP("Arch_PipeConnector", "Connector"),
            "Accel": "P, C",
            "ToolTip": QT_TRANSLATE_NOOP(
                "Arch_PipeConnector", "Creates a connector between 2 or 3 selected pipes"
            ),
        }

    def IsActive(self):

        v = hasattr(FreeCADGui.getMainWindow().getActiveWindow(), "getSceneGraph")
        return v

    def Activated(self):

        import Draft

        s = FreeCADGui.Selection.getSelection()
        if not (len(s) in [2, 3]):
            FreeCAD.Console.PrintError(
                translate("Arch", "Select exactly 2 or 3 pipe objects") + "\n"
            )
            return
        o = "["
        for obj in s:
            if Draft.getType(obj) != "Pipe":
                FreeCAD.Console.PrintError(translate("Arch", "Select only pipe objects") + "\n")
                return
            o += "FreeCAD.ActiveDocument." + obj.Name + ","
        o += "]"
        FreeCAD.ActiveDocument.openTransaction(translate("Arch", "Create Connector"))
        FreeCADGui.addModule("Arch")
        FreeCADGui.addModule("Draft")
        FreeCADGui.doCommand("obj = Arch.makePipeConnector(" + o + ")")
        FreeCADGui.doCommand("Draft.autogroup(obj)")
        FreeCAD.ActiveDocument.commitTransaction()
        FreeCAD.ActiveDocument.recompute()


class Arch_PipeGroupCommand:

    def GetCommands(self):
        return tuple(["Arch_Pipe", "Arch_PipeConnector"])

    def GetResources(self):
        return {
            "MenuText": QT_TRANSLATE_NOOP("Arch_PipeTools", "Pipe Tools"),
            "ToolTip": QT_TRANSLATE_NOOP("Arch_PipeTools", "Pipe tools"),
        }

    def IsActive(self):
        v = hasattr(FreeCADGui.getMainWindow().getActiveWindow(), "getSceneGraph")
        return v


FreeCADGui.addCommand("Arch_Pipe", Arch_Pipe())
FreeCADGui.addCommand("Arch_PipeConnector", Arch_PipeConnector())
FreeCADGui.addCommand("Arch_PipeTools", Arch_PipeGroupCommand())



# SPDX-License-Identifier: LGPL-2.1-or-later

# flake8: noqa E731
# ***************************************************************************
# *   Copyright (c) 2025 Samuel Abels <knipknap@gmail.com>                  *
# *                                                                         *
# *   This program is free software; you can redistribute it and/or modify  *
# *   it under the terms of the GNU Lesser General Public License (LGPL)    *
# *   as published by the Free Software Foundation; either version 2 of     *
# *   the License, or (at your option) any later version.                   *
# *   for detail see the LICENCE text file.                                 *
# *                                                                         *
# *   This program is distributed in the hope that it will be useful,       *
# *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
# *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
# *   GNU Library General Public License for more details.                  *
# *                                                                         *
# *   You should have received a copy of the GNU Library General Public     *
# *   License along with this program; if not, write to the Free Software   *
# *   Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  *
# *   USA                                                                   *
# *                                                                         *
# ***************************************************************************

"""Widget for browsing ToolBit assets with filtering and sorting."""

import yaml
from typing import List, Optional, cast, Sequence
from PySide import QtGui, QtCore
from PySide.QtGui import QApplication, QMessageBox, QMenu, QAction, QKeySequence, QDialog
from PySide.QtCore import QMimeData
import FreeCAD
import Path
from ...assets import AssetManager, AssetUri
from ..models.base import ToolBit
from ..serializers.yaml import YamlToolBitSerializer
from .toollist import ToolBitListWidget, CompactToolBitListWidget, ToolBitUriRole
from .editor import ToolBitEditor
from .util import natural_sort_key


Path.Log.setLevel(Path.Log.Level.INFO, Path.Log.thisModule())
Path.Log.trackModule(Path.Log.thisModule())


class ToolBitBrowserWidget(QtGui.QWidget):
    """
    A widget to browse, filter, and select ToolBit assets from the
    AssetManager, with sorting and batch insertion.
    """

    # Signal emitted when a tool is selected in the list
    toolSelected = QtCore.Signal(str)  # Emits ToolBit URI string
    # Signal emitted when a tool is requested for editing (e.g., double-click)
    itemDoubleClicked = QtCore.Signal(ToolBit)  # Emits ToolBit URI string

    # Debounce timer for search input
    _search_timer_interval = 300  # milliseconds

    def __init__(
        self,
        asset_manager: AssetManager,
        store: str = "local",
        parent=None,
        tool_no_factory=None,
        tool_fetcher=None,
        compact=False,
    ):
        super().__init__(parent)
        self._asset_manager = asset_manager
        self._tool_no_factory = tool_no_factory
        self._compact_mode = compact

        self._is_fetching = False
        self._store_name = store
        self._all_assets: Sequence[ToolBit] = []  # Store all fetched assets
        self._current_search = ""  # Track current search term
        self._sort_key = "tool_no" if tool_no_factory else "label"
        self._selected_uris: List[str] = []  # Track selected toolbit URIs

        # UI Elements
        self._search_edit = QtGui.QLineEdit()
        self._search_edit.setPlaceholderText("Search toolbits...")

        # Sorting dropdown
        self._sort_combo = QtGui.QComboBox()
        if self._tool_no_factory:
            self._sort_combo.addItem("Sort by Toolbit Number", "tool_no")
        self._sort_combo.addItem("Sort by Label", "label")
        self._sort_combo.setCurrentIndex(0)
        self._sort_combo.setVisible(self._tool_no_factory is not None)  # Hide if no tool_no_factory

        # Top layout for search and sort
        self._top_layout = QtGui.QHBoxLayout()
        self._top_layout.addWidget(self._search_edit, 3)  # Give search more space
        self._top_layout.addWidget(self._sort_combo, 1)

        if self._compact_mode:
            self._tool_list_widget = CompactToolBitListWidget(tool_no_factory=self._tool_no_factory)
        else:
            self._tool_list_widget = ToolBitListWidget(tool_no_factory=self._tool_no_factory)

        # Main layout
        layout = QtGui.QVBoxLayout(self)
        layout.addLayout(self._top_layout)
        layout.addWidget(self._tool_list_widget)

        # Connections
        self._search_timer = QtCore.QTimer(self)
        self._search_timer.setSingleShot(True)
        self._search_timer.setInterval(self._search_timer_interval)
        self._search_timer.timeout.connect(self._update_list)
        self._search_edit.textChanged.connect(self._search_timer.start)
        self._sort_combo.currentIndexChanged.connect(self._on_sort_changed)

        # Connect signals from the list widget
        self._tool_list_widget.itemDoubleClicked.connect(self._on_item_double_clicked)
        self._tool_list_widget.itemSelectionChanged.connect(self._on_item_selection_changed)

        # Connect list widget context menu request to browser handler
        self._tool_list_widget.setContextMenuPolicy(QtCore.Qt.CustomContextMenu)
        self._tool_list_widget.customContextMenuRequested.connect(self._show_context_menu)

        # Add keyboard shortcuts
        self._add_shortcuts()

        # Note that fetching of assets is done at showEvent(),
        # because we need to know the widget size to calculate the number
        # of items that need to be fetched.
        self.tool_fetcher = tool_fetcher or self._tool_fetcher

    def showEvent(self, event):
        """Handles the widget show event to trigger initial data fetch."""
        super().showEvent(event)
        # Fetch all assets the first time the widget is shown
        if not self._all_assets and not self._is_fetching:
            self.refresh()
        # Set focus to the search field
        self._search_edit.setFocus()

    def _tool_fetcher(self) -> Sequence[ToolBit]:
        return cast(
            List[ToolBit],
            self._asset_manager.fetch(
                asset_type="toolbit",
                depth=0,  # do not fetch dependencies (e.g. shape, icon)
                store=self._store_name,
            ),
        )

    def select_by_uri(self, uris: List[str]):
        if not uris:
            return

        # Select and scroll to the first toolbit
        is_first = True
        for i in range(self._tool_list_widget.count()):
            item = self._tool_list_widget.item(i)
            if item.data(ToolBitUriRole) in uris:
                self._tool_list_widget.setCurrentItem(item)
                if is_first:
                    # Scroll to the first selected item
                    is_first = False
                    self._tool_list_widget.scrollToItem(item)

    def refresh(self):
        """Fetches all ToolBit assets and stores them in memory, then updates the UI."""
        if self._is_fetching:
            return
        self._is_fetching = True
        try:
            self._all_assets = self.tool_fetcher()
        finally:
            self._is_fetching = False
        Path.Log.debug(f"Loaded {len(self._all_assets)} ToolBits.")

        self._sort_assets()
        self._update_list()

    def _sort_assets(self):
        """Sorts the in-memory assets based on the current sort key."""
        if self._sort_key == "label":
            self._all_assets.sort(key=lambda x: natural_sort_key(x.label))
        elif self._sort_key == "tool_no" and self._tool_no_factory:
            self._all_assets.sort(
                key=lambda x: int(self._tool_no_factory(x) or 0) if self._tool_no_factory else 0
            )

    def _matches_search(self, toolbit, search_term):
        """Checks if a ToolBit matches the search term."""
        search_term = search_term.lower()
        return search_term in toolbit.label.lower() or search_term in toolbit.summary.lower()

    def _update_list(self):
        """Updates the list widget based on current search and sort."""
        if self._is_fetching:
            return

        self._current_search = self._search_edit.text()
        filtered_assets = [
            asset
            for asset in self._all_assets
            if not self._current_search or self._matches_search(asset, self._current_search)
        ]

        # Collect current items in the list widget
        current_items = {}
        for i in range(self._tool_list_widget.count()):
            item = self._tool_list_widget.item(i)
            uri = item.data(ToolBitUriRole)
            if uri:
                current_items[uri] = item

        # Iterate through filtered assets and update the list widget
        for i, asset in enumerate(filtered_assets):
            uri = str(asset.get_uri())
            if uri in current_items:
                # Item exists, remove the old one and insert the new one
                item = current_items[uri]
                row = self._tool_list_widget.row(item)
                self._tool_list_widget.takeItem(row)
                self._tool_list_widget.insert_toolbit(i, asset)
                del current_items[uri]
            else:
                # Insert new item
                self._tool_list_widget.insert_toolbit(i, asset)

        # Remove items that are no longer in filtered_assets
        for uri, item in current_items.items():
            row = self._tool_list_widget.row(item)
            self._tool_list_widget.takeItem(row)

        # Restore selection and scroll to the selected item
        if self._selected_uris:
            first_selected_item = None
            for i in range(self._tool_list_widget.count()):
                item = self._tool_list_widget.item(i)
                uri = item.data(ToolBitUriRole)
                if uri in self._selected_uris:
                    item.setSelected(True)
                    if first_selected_item is None:
                        first_selected_item = item
            if first_selected_item:
                self._tool_list_widget.scrollToItem(first_selected_item)

        # Apply the filter to trigger highlighting in the list widget
        self._tool_list_widget.apply_filter(self._current_search)

    def set_sort_order(self, key: str):
        for i in range(self._sort_combo.count()):
            if self._sort_combo.itemData(i) == key:
                if self._sort_combo.currentIndex() != i:
                    self._sort_combo.setCurrentIndex(i)
                break
        else:
            return
        self._sort_key = key
        self._sort_assets()
        self._update_list()

    def _on_sort_changed(self):
        """Handles sort order change from the dropdown."""
        key = self._sort_combo.itemData(self._sort_combo.currentIndex())
        self.set_sort_order(key)

    def _on_item_double_clicked(self, item):
        """Handles double-click on a list item to request editing."""
        uri_string = item.data(ToolBitUriRole)
        if not uri_string:
            return
        try:
            toolbit = self._asset_manager.get(AssetUri(uri_string))
            if toolbit:
                self.itemDoubleClicked.emit(toolbit)
        except FileNotFoundError:
            # Handle missing/placeholder toolbits gracefully
            QMessageBox.warning(
                self,
                FreeCAD.Qt.translate("CAM", "Missing Toolbit"),
                FreeCAD.Qt.translate(
                    "CAM",
                    "This toolbit is missing from your local store. It may be a placeholder for a toolbit that was not found during library import.",
                ),
            )
        except Exception as e:
            # Handle other errors
            QMessageBox.critical(
                self,
                FreeCAD.Qt.translate("CAM", "Error"),
                FreeCAD.Qt.translate("CAM", f"Failed to load toolbit: {e}"),
            )

    def _on_item_selection_changed(self):
        """Emits toolSelected signal and tracks selected URIs."""
        selected_uris = self._tool_list_widget.get_selected_toolbit_uris()
        self._selected_uris = selected_uris
        if not selected_uris:
            return
        self.toolSelected.emit(selected_uris[0])

    def _get_first_selected_bit(self) -> Optional[ToolBit]:
        uris = self.get_selected_bit_uris()
        if not uris:
            return None
        uri_string = uris[0]
        return cast(ToolBit, self._asset_manager.get(AssetUri(uri_string)))

    def _on_edit_requested(self):
        """Opens the ToolBitEditor for the selected toolbit."""
        toolbit = self._get_first_selected_bit()
        if not toolbit:
            return

        # Open the editor for the selected toolbit
        editor = ToolBitEditor(toolbit)
        result = editor.show()
        if result != QDialog.Accepted:
            return

        # If the editor was closed with "OK", save the changes
        self._asset_manager.add(toolbit)
        Path.Log.info(f"Toolbit {toolbit.get_id()} saved.")
        self.refresh()
        self._update_list()

    def _add_shortcuts(self):
        """Adds keyboard shortcuts for common actions."""
        copy_action = QAction(self)
        copy_action.setShortcut(QKeySequence.Copy)
        copy_action.triggered.connect(self._on_copy_requested)
        self.addAction(copy_action)

        delete_action = QAction(self)
        delete_action.setShortcut(QKeySequence("Shift+Delete"))
        delete_action.triggered.connect(self._on_delete_requested)
        self.addAction(delete_action)

        edit_action = QAction(self)
        edit_action.setShortcut(QKeySequence("F2"))
        edit_action.triggered.connect(self._on_edit_requested)
        self.addAction(edit_action)

    def _create_base_context_menu(self):
        """Creates the base context menu with Edit, Copy, and Delete actions."""
        selected_items = self._tool_list_widget.selectedItems()
        has_selection = bool(selected_items)

        context_menu = QMenu(self)

        edit_action = context_menu.addAction("Edit", self._on_edit_requested)
        edit_action.setEnabled(has_selection)
        context_menu.addSeparator()
        action = context_menu.addAction("Copy", self._on_copy_requested)
        action.setShortcut(QKeySequence.Copy)
        action = context_menu.addAction("Delete from disk", self._on_delete_requested)
        action.setShortcut(QKeySequence("Shift+Delete"))

        return context_menu

    def _show_context_menu(self, position):
        """Shows the context menu at the given position."""
        context_menu = self._create_base_context_menu()
        context_menu.exec_(self._tool_list_widget.mapToGlobal(position))

    def _to_clipboard(
        self,
        uris: List[str],
        mode: str = "copy",
        extra_data: Optional[dict] = None,
    ):
        """Copies selected toolbits to the clipboard as YAML."""
        if not uris:
            return

        selected_bits = [cast(ToolBit, self._asset_manager.get(AssetUri(uri))) for uri in uris]
        selected_bits = [bit for bit in selected_bits if bit]  # Filter out None
        if not selected_bits:
            return

        # Serialize selected toolbits individually
        serialized_toolbits_data = []
        for toolbit in selected_bits:
            yaml_data = YamlToolBitSerializer.serialize(toolbit)
            serialized_toolbits_data.append(yaml_data.decode("utf-8"))

        # Create a dictionary to hold the operation type and serialized data
        clipboard_data_dict = {
            "operation": mode,
            "toolbits": serialized_toolbits_data,
        }

        # Include extra data if provided
        if extra_data:
            clipboard_data_dict.update(extra_data)

        # Serialize the dictionary to YAML
        clipboard_content_yaml = yaml.dump(clipboard_data_dict, default_flow_style=False)

        # Put the YAML data on the clipboard with a custom MIME type
        mime_data = QMimeData()
        mime_type = "application/x-freecad-toolbit-list-yaml"
        mime_data.setData(mime_type, clipboard_content_yaml.encode("utf-8"))

        # Put it in text format for pasting to text editors
        toolbit_list = [yaml.safe_load(d) for d in serialized_toolbits_data]
        mime_data.setText(yaml.dump(toolbit_list, default_flow_style=False))

        clipboard = QApplication.clipboard()
        clipboard.setMimeData(mime_data)

    def _on_copy_requested(self):
        """Copies selected toolbits to the clipboard as YAML."""
        uris = self.get_selected_bit_uris()
        self._to_clipboard(uris, mode="copy")

    def _on_delete_requested(self):
        """Deletes selected toolbits and removes them from all libraries."""
        Path.Log.debug("ToolBitBrowserWidget._on_delete_requested: Function entered.")
        uris = self.get_selected_bit_uris()
        if not uris:
            Path.Log.debug("_on_delete_requested: No URIs selected. Returning.")
            return

        # Ask for confirmation
        reply = QMessageBox.question(
            self,
            FreeCAD.Qt.translate("CAM", "Confirm Deletion"),
            FreeCAD.Qt.translate(
                "CAM",
                "Are you sure you want to delete the selected toolbit(s)? This is not reversible. The toolbits will be removed from disk and from all libraries that contain them.",
            ),
            QMessageBox.Yes | QMessageBox.No,
            QMessageBox.No,
        )

        if reply != QMessageBox.Yes:
            return

        deleted_count = 0
        libraries_modified = []  # Use list instead of set since Library objects aren't hashable

        for uri_string in uris:
            try:
                toolbit_uri = AssetUri(uri_string)

                # First, remove the toolbit from all libraries that contain it
                libraries_to_update = self._find_libraries_containing_toolbit(toolbit_uri)
                for library in libraries_to_update:
                    library.remove_bit_by_uri(uri_string)
                    if library not in libraries_modified:  # Avoid duplicates
                        libraries_modified.append(library)
                    Path.Log.info(
                        f"Removed toolbit {toolbit_uri.asset_id} from library {library.label}"
                    )

                # Then delete the toolbit file from disk
                self._asset_manager.delete(toolbit_uri)
                deleted_count += 1
                Path.Log.info(f"Deleted toolbit file {uri_string}")

            except Exception as e:
                Path.Log.error(f"Failed to delete toolbit {uri_string}: {e}")

        # Save all modified libraries
        for library in libraries_modified:
            try:
                self._asset_manager.add(library)
                Path.Log.info(f"Saved updated library {library.label}")
            except Exception as e:
                Path.Log.error(f"Failed to save library {library.label}: {e}")

        if deleted_count > 0:
            Path.Log.info(
                f"Deleted {deleted_count} toolbit(s) and updated {len(libraries_modified)} libraries."
            )
            self.refresh()

    def _find_libraries_containing_toolbit(self, toolbit_uri: AssetUri) -> List:
        """Find all libraries that contain the specified toolbit."""
        from ...library.models.library import Library

        libraries_with_toolbit = []
        try:
            # Get all libraries from the asset manager
            all_libraries = self._asset_manager.fetch("toolbitlibrary", store="local", depth=1)

            for library in all_libraries:
                if isinstance(library, Library):
                    # Check if this library contains the toolbit
                    for toolbit in library:
                        if toolbit.get_uri() == toolbit_uri:
                            libraries_with_toolbit.append(library)
                            break

        except Exception as e:
            Path.Log.error(f"Error finding libraries containing toolbit {toolbit_uri}: {e}")

        return libraries_with_toolbit

    def get_selected_bit_uris(self) -> List[str]:
        """
        Returns a list of URIs for the currently selected ToolBit items.
        Delegates to the underlying list widget.
        """
        return self._tool_list_widget.get_selected_toolbit_uris()

    def get_selected_bits(self) -> List[ToolBit]:
        """
        Returns a list of selected ToolBit objects.
        Retrieves the full ToolBit objects using the asset manager.
        """
        selected_bits = []
        selected_uris = self.get_selected_bit_uris()
        for uri_string in selected_uris:
            toolbit = self._asset_manager.get(AssetUri(uri_string))
            if toolbit:
                selected_bits.append(toolbit)
        return selected_bits
